{************************************}
{*** Jednotka pro 3D Grafiku ********}
{*** c Dalibor Martiek *************}
{*** 1.3.2005 ***********************}
{************************************}

unit Graph3d_2005;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

const  { ** promtac metody - projection types **}
       ptParallel = 0;          {kosohl axonometrie}
       ptOrtho    = 1;          {kolm axonometrie}
       ptCenter   = 2;          {stedov promtn}
       ptSphere   = 3;          {promtn na kulovou plochu}
type
  TPixel           = array [1..2]      of Integer;
  T2DPoint         = array [1..2]      of Double;
  T3DPoint         = array [1..4]      of Double;
  TArrayOf3DPoints = array [0..800]    of T3DPoint;
  TMatrix          = array [1..4,1..4] of Double;
  TVector          = array [1..4]      of Double;
  TBox             = record
                       A,B,C,D,Ac,Bc,Cc,Dc:T3DPoint;
                     end;
  TTerm = record
               Character:Char;
               Value    :Double;
             end;
  TExpression = array [0..255] of TTerm;

  TDraw3D = class(TForm)
    Panel: TPanel;
    Image: TImage;
    MemoryImage: TImage;

  private
    { Private declarations }

  public
    { Public declarations }
    {Barevn sloky}
    Red,Green,Blue   :Byte;
    {Typ promtn}
    ProjectionType   :Byte;
    {hly pozorovn}
    AlfaView,BetaView:Double;
    {Smr pohledu pozorovatele}
    DirectOfView     :TVector;
    {Smr. hly svtelnho zdroje}
    AlfaLight,BetaLight:Double;
    {Smr svtla}
    DirectOfLight     :TVector;
   {Vzdlenost pozorovatele}
    ObserverDistance :Double;
    {Msto pozorovn}
    ObserverPoint    :T3DPoint;
    {Hlavn paprsek}
    MajorRay         :TVector;
    {Velikost uivatelskch 2D jednotek ve svtovch souadnicch}
    Unit_X,Unit_Y    :Double;
    {Svtov souadnice uivatelskho 2D potku}
    O                :T2DPoint;
    {Body}
    A,B,C,D,Ac,Bc,Cc,Dc   :T3DPoint;
    {koeficienty prmtu 3D objektu do 2D}
    i1,i2,j1,j2,k1,k2,
    {uivatelsk rozmry kreslc 2D plochy}
    x1,x2,y1,y2,
    {funkn promnn a jejich kroky}
    x,y,z,hx,hy,hz,k,hk,
    {parametr pro parametrickou kivku, jeho interval a krok}
    t,t1,t2,ht,r,r1,r2,hr,s,s1,s2,hs,n,
    {argument pro polrn souadnice, jeho interval a krok}
    fi,fi1,fi2,hfi      :Double;

    {pole bod pro kivky}
    P,Q                  :TArrayOf3DPoints;

    {***** svtov souadnice 2D *****}

    {Nastaven velikosti obrazu}
    Procedure InitImage(Width,Height:Integer);
    {Nastaven velikosti pamti pro obraz}
    Procedure InitMemoryImage(Width,Height:Integer);
    {nastaven barvy z barevnch sloek}
    Procedure ClearImage(Red,Green,Blue:byte);
    {Nastaven uivatelskch rozmr kreslc plochy}
    procedure Scale(x1,x2,y1,y2:Double);
    {Pepoet prvn uivatelsk souadnice na svtovou}
    function  XCoor(x:Double):integer;
    {Pepoet druh uivatelsk souadnice na svtovou}
    function  YCoor(y:Double):integer;
    {Pepoet prvn svtov souadnice na uivatelskou}
    function  InvXCoor(x:Integer):Double;
    {Pepoet druh svtov souadnice na uivatelskou}
    function  InvYCoor(y:Integer):Double;
    {Sestroj bod ve svtovch souadnicch zadanou barvou}
    procedure PutPixel(X:TPixel;Red,Green,Blue:byte);
    {Zjist barvu bodu zadanho ve svtovch souadnicch}
    Procedure GetPixel(X:TPixel;var Red,Green,Blue:byte);
    {Sestroj bod ve svtovch souadnicch zadan prhlednosti zadanou barvou}
    Procedure PutTransparentPixel(X:TPixel;Red,Green,Blue:Byte;Transparency:Double);
    {Sestroj seku ve svtovch souadnicch zadanou barvou}
    procedure Line2D(X,Y:TPixel;Red,Green,Blue:byte);
    {Sestroj seku v pamti ve svtovch souadnicch}
    procedure MemoryLine2D(X,Y:TPixel;Red,Green,Blue:byte);
    {Sestroj seku interpolac barvy ve svtovch souadnicch}
    procedure InterpolLine2D(X,Y:TPixel;RedX,GreenX,BlueX,
                                        RedY,GreenY,BlueY:byte);
    {Sestroj seku interpolac barvy pamti ve svtovch souadnicch}
    procedure MemoryInterpolLine2D(X,Y:TPixel;RedX,GreenX,BlueX,
                                              RedY,GreenY,BlueY:byte);
    {Sestroj trojhelnk ve svtovch souadnicch zadanou barvou}
    procedure Triangle2D(X,Y,Z:TPixel;Red,Green,Blue:byte);
    {Sestroj trojhelnk ve svtovch souadnicch vyplnn zadanou barvou}
    procedure FillTriangle2D(X,Y,Z:TPixel;Red,Green,Blue:byte);
   {Sestroj  drsn trojhelnk ve svtovch souadnicch vyplnn zadanou barvou}
    Procedure FillRoughTriangle2D(X,Y,Z:TPixel;Red,Green,Blue,Roughness:Byte;ColorDisp:Boolean);
     {Sestroj prhledn trojhelnk ve svtovch souadnicch vyplnn zadanou barvou}
    Procedure FillTranspTriangle2D(X,Y,Z:TPixel;Red,Green,Blue:Byte;Transparency:Double);
    {Sestroj trojhelnk interpolac barvy ve vrcholech ve svtovch souadnicch}

    procedure FillInterpolTriangle2D(X,Y,Z:TPixel;RedX,GreenX,BlueX,
                                                  RedY,GreenY,BlueY,
                                                  RedZ,GreenZ,BlueZ:byte);
   {Sestroj prhledn trojhelnk vyplnn interpolac barvy ve vrcholech}
   procedure FillInterpolTranspTriangle2D
                        (X,Y,Z:TPixel;RedX,GreenX,BlueX,
                                      RedY,GreenY,BlueY,
                                      RedZ,GreenZ,BlueZ:byte;Transparency:Double);

    {***** Promtac metody ********}

    {Prmt bodu z 3D do prmtny}
    procedure Projection(X:T3DPoint;var Xp:T2DPoint);
    {Prsek promtacho paprsku s rovinnou prmtnou}
    Procedure PlaneIntersect(X:T3DPoint;var pX:T3DPoint);
    {Prsek promtacho paprsku s kulovou prmtnou}
    Procedure SphereIntersect(X:T3DPoint;var pX:T3DPoint);
    {Nastaven smru pohledu pro paraleln projekci}
    procedure SetOrthoAxonProjection(Alfa,Beta:Double);
    {Nastaven stedu promtn pro centrln projekci}
    procedure SetCenterProjection(Alfa,Beta,ObserverDistance:Double;var ObserverPoint:T3DPoint);
    {Nastaven sfrick projekce}
    procedure SetSphereProjection(Alfa,Beta,ObserverDistance:Double;var ObserverPoint:T3DPoint);
    {Nastaven smru svtla}
    procedure SetDirectOfLight(AlfaLight,BetaLight:Double;var DirectOfLight:TVector);

    {***** Pomocn vpoty - ScanLine  3D *****}

    {Vektor uren dvma body}
    Procedure SetVector(X,Y:T3DPoint;var v:TVector);
    {Skalrn souin}
    Function ScalarProduct(u,v:TVector):Double;
    {Vektorov souin}
    Procedure VectorProduct(r,s:TVector;var t:TVector);
    {Velikost vektoru}
    Function Norm(u:TVector):Double;
    {Kosinus hlu vektor}
    function CosAngle(r,s:TVector):Double;

    {***** Konstrukce objekt v uivatelskch souadnicch - ScanLine  3D *****}

    {Sestroj bod na uivatelskch souadnicch zadanou barvou}
    Procedure Put3DPoint(X:T3DPoint;Red,Green,Blue:byte);
    {Sestroj kek na uivatelskch souadnicch zadanou barvou}
    Procedure Cross(X:T3DPoint;Size,Red,Green,Blue:byte);
    {Sestroj seku v uivatelskch souadnicch zadanou barvou}
    procedure Line(A,B:T3DPoint;Red,Green,Blue:byte);
     {Sestroj polygon prochzenc prvnmi n body zadanho pole zadanou barvou}
    procedure PolyLine(X:TArrayOf3DPoints;n:Word;Red,Green,Blue:byte);
    {Sestroj trojhelnk zadanou barvou}
    procedure Triangle(A,B,C:T3DPoint;Red,Green,Blue:byte);
    {Sestroj trojhelnk vyplnn zadanou barvou}
    procedure FillTriangle(A,B,C:T3DPoint;Red,Green,Blue:byte);
    {Sestroj drsn trojhelnk vyplnn zadanou barvou}
    procedure FillRoughTriangle(A,B,C:T3DPoint;Red,Green,Blue,Roughness:byte;ColorDisp:boolean);
    {Sestroj prhledn trojhelnk vyplnn zadanou barvou}
    procedure FillTranspTriangle(A,B,C:T3DPoint;Red,Green,Blue:byte;Transparency:Double);
    {Sestroj trojhelnk vyplnn interpolac barvy ve vrcholech}
    procedure FillInterpolTriangle(A,B,C:T3DPoint;RedA,GreenA,BlueA,
                                                    RedB,GreenB,BlueB,
                                                    RedC,GreenC,BlueC:byte);
    {Sestroj prhledn trojhelnk vyplnn interpolac barvy ve vrcholech}
    procedure FillInterpolTranspTriangle
                        (A,B,C:T3DPoint;RedA,GreenA,BlueA,
                                        RedB,GreenB,BlueB,
                                        RedC,GreenC,BlueC:byte;Transparency:Double);

    {***** uivatelsk souadnice - Canvas  3D  *****}

    {Sestroj seku v uivatelskch souadnicch zadanou barvou}
    procedure cvLine(A,B:T3DPoint;Red,Green,Blue:byte);
    {Sestroj polygon prochzenc prvnmi n body zadanho pole zadanou barvou}
    procedure cvPolyLine(X:TArrayOf3DPoints;n:Word;Red,Green,Blue:byte);
    {Sestroj hrany estistnu urenho vrcholy zadanou barvou}
    procedure cvBox(Box:TBox;Red,Green,Blue:byte);

   {**** nastaven matic zakladnich tansformac *****}

   {transformace bodu X matic Z do bodu X'}
    procedure TransOfPoint(X:T3DPoint;Z:TMatrix;var Xc:T3DPoint);
   {transformace trojhelnka ABC matic Z do trojhelnka A'B'C'}
    Procedure TransOfTriangle(A,B,C:T3DPoint;Z:TMatrix;var Ac,Bc,Cc:T3DPoint);

    procedure TransOfBox(Box1:TBox;Z:TMatrix;var Box2:TBox);
   {**** nastaven matic slozench transformac **********}
    {nsoben matic}
    Procedure Multip(M,N:TMatrix;var Result:TMatrix);
    {posunut o vektor v}
    Procedure Transl(V:TVector;var M:TMatrix);
   {otoen o hel alfa kolem osy X}
    Procedure RotX(Alfa:Double;var M:TMatrix);
   {otoen o hel alfa kolem osy Y}
    Procedure RotY(Beta:Double;var M:TMatrix);
   {otoen o hel alfa kolem osy Z}
    Procedure RotZ(Gamma:Double;var M:TMatrix);
   {stedov soumrnost se stedem v potku}
    procedure AxisSymX(var M:TMatrix);
    procedure AxisSymY(var M:TMatrix);
    procedure AxisSymZ(var M:TMatrix);
    procedure PlaneSymXY(var M:TMatrix);
    procedure PlaneSymYZ(var M:TMatrix);
    procedure PlaneSymXZ(var M:TMatrix);
    procedure OriginSym(var M:TMatrix);
    {rotace kolem obecn osy}
    procedure GenRot(A,B:T3DPoint;Alfa:Double;var M:TMatrix);

    Function PreCalc(Vstup:String;var Errorreport:String):TExpression;
    Function PostCalc(Vstup:TExpression;var ErrorReport:String):Double;
    Function Calc(Vstup:String;var ErrorReport:String):Double;

  end;  {TDraw3D Declaration}

Const Max     =255;                                    {maximalni dlka etzce}
      NoOfFunct=14;
      Digits  =['0'..'9','.'];
      BinOper =['^','*','/','+','-'];                        {binrn opertory}
      UnOper  =[#199..#211];                                  {unrn opertory}
      Functions :array [1..NoOfFunct] of string =
                    ('E','ABS','SIN','SQR','SQRT','COS','ARCTAN','ARCSIN','ARCCOS',
                   'EXP','LN','TAN','FA','PI');
      Letters = ['E','A','B','S','I','N','Q','R','T','C','O','K','N',
                                     'X','P','L','F','Y','Z'];
      Variables = ['K','N','R','S','T','X','Y','Z'];
      PractZero  = 1e-15;

var  Draw3D: TDraw3D;

implementation

{$R *.DFM}

Procedure TDraw3D.InitImage(Width,Height:Integer);
begin
      Panel.Align:=alNone;
      Panel.Width:=Width;
      Panel.Height:=Height;
      Image.Align:=alNone;
      Image.Picture.Bitmap.Create;
      Image.Picture.Bitmap.PixelFormat:=pf24bit;
      Image.Picture.Bitmap.Width:=Width;
      Image.Picture.Bitmap.Height:=Height;
      Image.Width:=Width;
      Image.Height:=Height;
      Image.Visible:=True;
end;

Procedure TDraw3D.InitMemoryImage(Width,Height:Integer);
begin
      MemoryImage.Align:=alNone;
      MemoryImage.Picture.Bitmap.PixelFormat:=pf24bit;
      MemoryImage.Picture.Bitmap.Width:=Width;
      MemoryImage.Picture.Bitmap.Height:=Height;
      MemoryImage.Width:=Width;
      MemoryImage.Height:=Height;
      MemoryImage.Visible:=False;
end;

Procedure TDraw3D.ClearImage(Red,Green,Blue:Byte);
var i,j,Adr:Integer;
    ScanRow: PByteArray;
begin
  With Image do
  for j:=0 to Height-1 do
  begin
    ScanRow:=Picture.Bitmap.ScanLine[j];
    for i:=0 to Width-1 do
    begin
      Adr:=3*i;
      ScanRow[Adr]:=Blue;
      ScanRow[succ(Adr)]:=Green;
      ScanRow[succ(succ(Adr))]:=Red;
    end;
  end;
end;

Procedure TDraw3D.Scale(x1,x2,y1,y2:Double);
  begin
     Unit_X:= Image.Width/(x2-x1);
     Unit_Y:= Image.Height/(y2-y1);
     O[1]  := -x1*Unit_X;
     O[2]  :=  y2*Unit_Y;
  end;

function  TDraw3D.XCoor(x:Double):integer;
  begin
    XCoor:=Round(O[1]+Unit_X*x);
  end;

function  TDraw3D.YCoor(y:Double):integer;
  begin
     YCoor:=Round(O[2]-Unit_Y*y);
  end;

 function  TDraw3D.InvXCoor(x:Integer):Double;
  begin
    InvXCoor:=(x-O[1])/Unit_X;
  end;

  function  TDraw3D.InvYCoor(y:Integer):Double;
  begin
     InvYCoor:=(O[2]-y)/Unit_Y;
  end;

procedure TDraw3D.PutPixel(X:TPixel;Red,Green,Blue:Byte);
  var Adr:Integer;
      ScanRow: PByteArray;
  begin
      ScanRow:=Image.Picture.Bitmap.ScanLine[X[2]];
      Adr:=3*X[1];
      ScanRow[Adr]:=Blue;
      ScanRow[succ(Adr)]:=Green;
      ScanRow[succ(succ(Adr))]:=Red;
  end;

procedure TDraw3D.GetPixel(X:TPixel;var Red,Green,Blue:Byte);
  var Adr:Integer;
      ScanRow: PByteArray;
  begin
      ScanRow:=Image.Picture.Bitmap.ScanLine[X[2]];
      Adr:=3*X[1];
      Blue:=ScanRow[Adr];
      Green:=ScanRow[succ(Adr)];
      Red:=ScanRow[succ(succ(Adr))];
  end;

Procedure TDraw3D.PutTransparentPixel(X:TPixel;Red,Green,Blue:Byte;Transparency:Double);
var R,G,B:Byte;
begin
  GetPixel(X,R,G,B);
  R:=Round(Transparency*R+(1-Transparency)*Red);
  G:=Round(Transparency*G+(1-Transparency)*Green);
  B:=Round(Transparency*B+(1-Transparency)*Blue);
  PutPixel(X,R,G,B);
end;

procedure TDraw3D.Line2D(X,Y:TPixel;Red,Green,Blue:Byte);
  var p,p1,p2,Dx,Dy,i       :Integer;
      M                     :TPixel;
      Position              :Byte;
      Adr                   :Integer;
      ScanRow: PByteArray;
  Begin
  With Image do
  if (X[1]<0) or (X[1]>=Width) or (X[2]<0) or (X[2]>=Height) or
     (Y[1]<0) or (Y[1]>=Width) or (Y[2]<0) or (Y[2]>=Height) then exit;
  Dx:=Y[1]-X[1];
  if Dx<0 then
       begin
             p:=X[1];X[1]:=Y[1];Y[1]:=p;
             p:=X[2];X[2]:=Y[2];Y[2]:=p;
             Dx:=-Dx;
       end;
  Dy:=Y[2]-X[2];M[1]:=X[1];M[2]:=X[2];
  ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
  Adr:=3*M[1];
  ScanRow[Adr]:=Blue;
  ScanRow[succ(Adr)]:=Green;
  ScanRow[succ(succ(Adr))]:=Red;
  if Dx=0 then
           if Dy>0 then Position:=1
                   else Position:=2
          else
           if Dy=0 then Position:=9
            else
            if Dy=Dx then Position:=3
                    else
                     if Dy=-Dx then Position:=4
                               else
                               if Dy>0 then
                                       if Dy>Dx then Position:=5
                                                 else Position:=6
                                        else
                                         if -Dy>Dx then Position:=7
                                                   else Position:=8;
 Case Position of
  1: For i:=succ(X[2]) to Y[2] do
            begin
              ScanRow:=Image.Picture.Bitmap.ScanLine[i];
              Adr:=3*M[1];
              ScanRow[Adr]:=Blue;
              ScanRow[succ(Adr)]:=Green;
              ScanRow[succ(succ(Adr))]:=Red;
            end;
  2: For i:=pred(X[2]) downto Y[2] do
            begin
              ScanRow:=Image.Picture.Bitmap.ScanLine[i];
              Adr:=3*M[1];
              ScanRow[Adr]:=Blue;
              ScanRow[succ(Adr)]:=Green;
              ScanRow[succ(succ(Adr))]:=Red;
            end;
  3: For i:=succ(X[1]) to Y[1] do
            begin
              inc(M[2]);
              ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
              Adr:=3*i;
              ScanRow[Adr]:=Blue;
              ScanRow[succ(Adr)]:=Green;
              ScanRow[succ(succ(Adr))]:=Red;
            end;
  4: For i:=succ(X[1]) to Y[1] do
           begin
             dec(M[2]);
             ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
             Adr:=3*i;
             ScanRow[Adr]:=Blue;
             ScanRow[succ(Adr)]:=Green;
             ScanRow[succ(succ(Adr))]:=Red;
           end;
  5: begin
       p:=Dx;Dx:=Dy;Dy:=p;
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;
       While M[2]<Y[2] do
       begin
        Inc(M[2]);
        if p>0 then begin
                       Inc(M[1]);p:=p+p2;
                    end
                    else       p:=p+p1;
        ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
        Adr:=3*M[1];
        ScanRow[Adr]:=Blue;
        ScanRow[succ(Adr)]:=Green;
        ScanRow[succ(succ(Adr))]:=Red;
      end;
     end;
  6: begin
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
       While M[1]<Y[1] do
       begin
        inc(M[1]);
        if p>0 then begin
                      inc(M[2]);p:=p+p2;
                      ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
                    end
                    else  p:=p+p1;
        Adr:=3*M[1];
        ScanRow[Adr]:=Blue;
        ScanRow[succ(Adr)]:=Green;
        ScanRow[succ(succ(Adr))]:=Red;
       end;
      end;
  7: begin
       p:=Dx;Dx:=Dy;Dy:=-p;
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;
       While M[2]>Y[2] do
       begin
        dec(M[2]);
        if p<0 then begin
                     inc(M[1]);p:=p+p2;
                    end
                    else       p:=p+p1;
        ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
        Adr:=3*M[1];
        ScanRow[Adr]:=Blue;
        ScanRow[succ(Adr)]:=Green;
        ScanRow[succ(succ(Adr))]:=Red;
      end;
     end;
  8: begin
       Dy:=-Dy;ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;
       While M[1]<Y[1] do
       begin
        inc(M[1]);
        if p>0 then begin
                      dec(M[2]);
                      p:=p+p2;
                      ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
                    end
                    else  p:=p+p1;
        Adr:=3*M[1];
        ScanRow[Adr]:=Blue;
        ScanRow[succ(Adr)]:=Green;
        ScanRow[succ(succ(Adr))]:=Red;
      end;
     end;
   9:begin
      ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
      For i:=succ(X[1]) to Y[1] do
      begin
       Adr:=3*i;
       ScanRow[Adr]:=Blue;
       ScanRow[succ(Adr)]:=Green;
       ScanRow[succ(succ(Adr))]:=Red;
      end;
    end;
  end;
end;      {Line2D}

procedure TDraw3D.MemoryLine2D(X,Y:TPixel;Red,Green,Blue:Byte);
  var p,p1,p2,Dx,Dy,i       :Integer;
      M                     :TPixel;
      Position              :Byte;
      Adr                   :Integer;
      ScanRow: PByteArray;
  Begin
  With MemoryImage do
  if (X[1]<0) or (X[1]>=Width) or (X[2]<0) or (X[2]>=Height) or
     (Y[1]<0) or (Y[1]>=Width) or (Y[2]<0) or (Y[2]>=Height) then exit;
  Dx:=Y[1]-X[1];
  if Dx<0 then
       begin
             p:=X[1];X[1]:=Y[1];Y[1]:=p;
             p:=X[2];X[2]:=Y[2];Y[2]:=p;
             Dx:=-Dx;
       end;
  Dy:=Y[2]-X[2];M[1]:=X[1];M[2]:=X[2];
  ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
  Adr:=3*M[1];
  ScanRow[Adr]:=Blue;
  ScanRow[succ(Adr)]:=Green;
  ScanRow[succ(succ(Adr))]:=Red;
  if Dx=0 then
           if Dy>0 then Position:=1
                   else Position:=2
          else
           if Dy=0 then Position:=9
            else
            if Dy=Dx then Position:=3
                    else
                     if Dy=-Dx then Position:=4
                               else
                               if Dy>0 then
                                       if Dy>Dx then Position:=5
                                                 else Position:=6
                                        else
                                         if -Dy>Dx then Position:=7
                                                   else Position:=8;
 Case Position of
  1: For i:=succ(X[2]) to Y[2] do
            begin
              ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[i];
              Adr:=3*M[1];
              ScanRow[Adr]:=Blue;
              ScanRow[succ(Adr)]:=Green;
              ScanRow[succ(succ(Adr))]:=Red;
            end;
  2: For i:=pred(X[2]) downto Y[2] do
            begin
              ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[i];
              Adr:=3*M[1];
              ScanRow[Adr]:=Blue;
              ScanRow[succ(Adr)]:=Green;
              ScanRow[succ(succ(Adr))]:=Red;
            end;
  3: For i:=succ(X[1]) to Y[1] do
            begin
              inc(M[2]);
              ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
              Adr:=3*i;
              ScanRow[Adr]:=Blue;
              ScanRow[succ(Adr)]:=Green;
              ScanRow[succ(succ(Adr))]:=Red;
            end;
  4: For i:=succ(X[1]) to Y[1] do
           begin
             dec(M[2]);
             ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
             Adr:=3*i;
             ScanRow[Adr]:=Blue;
             ScanRow[succ(Adr)]:=Green;
             ScanRow[succ(succ(Adr))]:=Red;
           end;
  5: begin
       p:=Dx;Dx:=Dy;Dy:=p;
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;
       While M[2]<Y[2] do
       begin
        Inc(M[2]);
        if p>0 then begin
                       Inc(M[1]);p:=p+p2;
                    end
                    else       p:=p+p1;
        ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
        Adr:=3*M[1];
        ScanRow[Adr]:=Blue;
        ScanRow[succ(Adr)]:=Green;
        ScanRow[succ(succ(Adr))]:=Red;
      end;
     end;
  6: begin
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;
       ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
       While M[1]<Y[1] do
       begin
        inc(M[1]);
        if p>0 then begin
                      inc(M[2]);p:=p+p2;
                      ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
                    end
                    else  p:=p+p1;
        Adr:=3*M[1];
        ScanRow[Adr]:=Blue;
        ScanRow[succ(Adr)]:=Green;
        ScanRow[succ(succ(Adr))]:=Red;
       end;
      end;
  7: begin
       p:=Dx;Dx:=Dy;Dy:=-p;
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;
       While M[2]>Y[2] do
       begin
        dec(M[2]);
        if p<0 then begin
                     inc(M[1]);p:=p+p2;
                    end
                    else       p:=p+p1;
        ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
        Adr:=3*M[1];
        ScanRow[Adr]:=Blue;
        ScanRow[succ(Adr)]:=Green;
        ScanRow[succ(succ(Adr))]:=Red;
      end;
     end;
  8: begin
       Dy:=-Dy;ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;
       While M[1]<Y[1] do
       begin
        inc(M[1]);
        if p>0 then begin
                      dec(M[2]);
                      p:=p+p2;
                      ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
                    end
                    else  p:=p+p1;
        Adr:=3*M[1];
        ScanRow[Adr]:=Blue;
        ScanRow[succ(Adr)]:=Green;
        ScanRow[succ(succ(Adr))]:=Red;
      end;
     end;
   9:begin
      ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
      For i:=succ(X[1]) to Y[1] do
      begin
       Adr:=3*i;
       ScanRow[Adr]:=Blue;
       ScanRow[succ(Adr)]:=Green;
       ScanRow[succ(succ(Adr))]:=Red;
      end;
    end;
  end;
end;      {MemoryLine2D}

procedure TDraw3D.InterpolLine2D(X,Y:TPixel;RedX,GreenX,BlueX,RedY,GreenY,BlueY:Byte);
  var p,p1,p2,Dx,Dy,i       :Integer;
      M                     :TPixel;
      Position              :Byte;
      Adr                   :Integer;
      StepR,StepG,StepB,R,G,B:Single;
      ScanRow: PByteArray;
  Begin
  With Image do
  if (X[1]<0) or (X[1]>=Width) or (X[2]<0) or (X[2]>=Height) or
     (Y[1]<0) or (Y[1]>=Width) or (Y[2]<0) or (Y[2]>=Height) then exit;
  Dx:=Y[1]-X[1];
  if Dx<0 then
       begin
             p:=X[1];X[1]:=Y[1];Y[1]:=p;
             p:=X[2];X[2]:=Y[2];Y[2]:=p;
             Dx:=-Dx;
             Position:=RedX;  RedX:=RedY;    RedY:=Position;
             Position:=GreenX;GreenX:=GreenY;GreenY:=Position;
             Position:=BlueX; BlueX:=BlueY;  BlueY:=Position;
       end;
  Dy:=Y[2]-X[2];M[1]:=X[1];M[2]:=X[2];
  ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
  Adr:=3*M[1];
  ScanRow[Adr]:=BlueX;
  ScanRow[succ(Adr)]:=GreenX;
  ScanRow[succ(succ(Adr))]:=RedX;
  if (Dx=0) and (Dy=0) then exit;
  if abs(Dx)>abs(Dy) then begin
                 StepR:=(RedY-RedX)/DX;
                 StepG:=(GreenY-GreenX)/DX;
                 StepB:=(BlueY-BlueX)/DX;
                end
           else begin
                 StepR:=(RedY-RedX)/abs(DY);
                 StepG:=(GreenY-GreenX)/abs(DY);
                 StepB:=(BlueY-BlueX)/abs(DY);
                end;
  if Dx=0 then
           if Dy>0 then Position:=1
                   else Position:=2
          else
           if Dy=0 then Position:=9
            else
            if Dy=Dx then Position:=3
                    else
                     if Dy=-Dx then Position:=4
                               else
                               if Dy>0 then
                                       if Dy>Dx then Position:=5
                                                 else Position:=6
                                        else
                                         if -Dy>Dx then Position:=7
                                                   else Position:=8;
 R:=RedX;G:=GreenX;B:=BlueX;
 Case Position of
  1:   For i:=succ(X[2]) to Y[2] do
            begin
              R:=R+StepR;G:=G+StepG;B:=B+StepB;
              ScanRow:=Image.Picture.Bitmap.ScanLine[i];
              Adr:=3*M[1];
              ScanRow[Adr]:=Round(B);
              ScanRow[succ(Adr)]:=Round(G);
              ScanRow[succ(succ(Adr))]:=Round(R);
            end;
 2:  For i:=pred(X[2]) downto Y[2] do
            begin
              R:=R+StepR;G:=G+StepG;B:=B+StepB;
              ScanRow:=Image.Picture.Bitmap.ScanLine[i];
              Adr:=3*M[1];
              ScanRow[Adr]:=Round(B);
              ScanRow[succ(Adr)]:=Round(G);
              ScanRow[succ(succ(Adr))]:=Round(R);
            end;
  3: For i:=succ(X[1]) to Y[1] do
            begin
              R:=R+StepR;G:=G+StepG;B:=B+StepB;
              inc(M[2]);
              ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
              Adr:=3*i;
              ScanRow[Adr]:=Round(B);
              ScanRow[succ(Adr)]:=Round(G);
              ScanRow[succ(succ(Adr))]:=Round(R);
            end;
  4: For i:=succ(X[1]) to Y[1] do
           begin
             R:=R+StepR;G:=G+StepG;B:=B+StepB;
             dec(M[2]);
             ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
             Adr:=3*i;
             ScanRow[Adr]:=Round(B);
             ScanRow[succ(Adr)]:=Round(G);
             ScanRow[succ(succ(Adr))]:=Round(R);
           end;
  5: begin
       p:=Dx;Dx:=Dy;Dy:=p;
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;
       While M[2]<Y[2] do
       begin
        R:=R+StepR;G:=G+StepG;B:=B+StepB;
        Inc(M[2]);
        if p>0 then begin
                       Inc(M[1]);p:=p+p2;
                    end
                    else       p:=p+p1;
        ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
        Adr:=3*M[1];
        ScanRow[Adr]:=Round(B);
        ScanRow[succ(Adr)]:=Round(G);
        ScanRow[succ(succ(Adr))]:=Round(R);
      end;
     end;
  6: begin
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;
       ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
       While M[1]<Y[1] do
       begin
        R:=R+StepR;G:=G+StepG;B:=B+StepB;
        inc(M[1]);
        if p>0 then begin
                      inc(M[2]);p:=p+p2;
                      ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
                    end
                    else  p:=p+p1;
        Adr:=3*M[1];
        ScanRow[Adr]:=Round(B);
        ScanRow[succ(Adr)]:=Round(G);
        ScanRow[succ(succ(Adr))]:=Round(R);
       end;
      end;
  7: begin
       p:=Dx;Dx:=Dy;Dy:=-p;
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;
       While M[2]>Y[2] do
       begin
        R:=R+StepR;G:=G+StepG;B:=B+StepB;
        dec(M[2]);
        if p<0 then begin
                     inc(M[1]);p:=p+p2;
                    end
                    else       p:=p+p1;
        ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
        Adr:=3*M[1];
        ScanRow[Adr]:=Round(B);
        ScanRow[succ(Adr)]:=Round(G);
        ScanRow[succ(succ(Adr))]:=Round(R);
      end;
     end;
  8: begin
       Dy:=-Dy;ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;
       While M[1]<Y[1] do
       begin
        R:=R+StepR;G:=G+StepG;B:=B+StepB;
        inc(M[1]);
        if p>0 then begin
                      dec(M[2]);
                      p:=p+p2;
                      ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
                    end
                    else  p:=p+p1;
        Adr:=3*M[1];
        ScanRow[Adr]:=Round(B);
        ScanRow[succ(Adr)]:=Round(G);
        ScanRow[succ(succ(Adr))]:=Round(R);
      end;
     end;
   9:begin
      ScanRow:=Image.Picture.Bitmap.ScanLine[M[2]];
      For i:=succ(X[1]) to Y[1] do
      begin
       R:=R+StepR;G:=G+StepG;B:=B+StepB;
       Adr:=3*i;
       ScanRow[Adr]:=Round(B);
       ScanRow[succ(Adr)]:=Round(G);
       ScanRow[succ(succ(Adr))]:=Round(R);
      end;
    end;
  end;
end;       {InterpolLine2D}

procedure TDraw3D.MemoryInterpolLine2D(X,Y:TPixel;RedX,GreenX,BlueX,RedY,GreenY,BlueY:Byte);
  var p,p1,p2,Dx,Dy,i       :Integer;
      M                     :TPixel;
      Position              :Byte;
      Adr                   :Integer;
      StepR,StepG,StepB,R,G,B:Single;
      ScanRow:PByteArray;
  Begin
  if (Y[1]=667) and (Y[2]=1217) and (X[1]=667) and (X[2]=1325)
               then
                 Adr:=0;
  With Image do
  if (X[1]<0) or (X[1]>=Width) or (X[2]<0) or (X[2]>=Height) or
     (Y[1]<0) or (Y[1]>=Width) or (Y[2]<0) or (Y[2]>=Height) then exit;
  Dx:=Y[1]-X[1];
  if Dx<0 then
       begin
             p:=X[1];X[1]:=Y[1];Y[1]:=p;
             p:=X[2];X[2]:=Y[2];Y[2]:=p;
             Dx:=-Dx;
             Position:=RedX;  RedX:=RedY;    RedY:=Position;
             Position:=GreenX;GreenX:=GreenY;GreenY:=Position;
             Position:=BlueX; BlueX:=BlueY;  BlueY:=Position;
       end;
  Dy:=Y[2]-X[2];M[1]:=X[1];M[2]:=X[2];
  ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
  Adr:=3*M[1];
  ScanRow[Adr]:=BlueX;
  ScanRow[succ(Adr)]:=GreenX;
  ScanRow[succ(succ(Adr))]:=RedX;
  if (Dx=0) and (Dy=0) then exit;                                {jedin bod}
  {if (Dx=0) then
              Adr:=succ(Adr)-1;}
  if abs(Dx)>abs(Dy) then begin
                 StepR:=(RedY-RedX)/DX;
                 StepG:=(GreenY-GreenX)/DX;
                 StepB:=(BlueY-BlueX)/DX;
                end
           else begin
                 StepR:=(RedY-RedX)/abs(DY);
                 StepG:=(GreenY-GreenX)/abs(DY);
                 StepB:=(BlueY-BlueX)/Abs(DY);
                end;
  if Dx=0 then
           if Dy>0 then Position:=1                           {svisle nahoru}
                   else Position:=2                             {svisle dolu}
          else
           if Dy=0 then Position:=3                           {zleva doprava}
            else
            if Dy=Dx then Position:=4                                   {k=1}
                    else
                     if Dy=-Dx then Position:=5                        {k=-1}
                               else
                               if Dy>0 then
                                       if Dy>Dx then Position:=6        {k>1}
                                                else Position:=7      {0<k<1}
                                        else
                                         if -Dy>Dx then Position:=8    {k<-1}
                                                   else Position:=9; {-1<k<0}
Case Position of
  {svisle nahoru}
  1: begin
     Adr:=3*M[1];R:=RedX;G:=GreenX;B:=BlueX;
     For i:=succ(X[2]) to Y[2] do
            begin
              R:=R+StepR;G:=G+StepG;B:=B+StepB;
              ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[i];
              ScanRow[Adr]:=Round(B);
              ScanRow[succ(Adr)]:=Round(G);
              ScanRow[succ(succ(Adr))]:=Round(R);
            end;
     end;
  {svisle dol}
  2: begin
      Adr:=3*M[1];R:=RedX;G:=GreenX;B:=BlueX;
      For i:=pred(X[2]) downto Y[2] do
            begin
              R:=R+StepR;G:=G+StepG;B:=B+StepB;
              ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[i];
              ScanRow[Adr]:=Round(B);
              ScanRow[succ(Adr)]:=Round(G);
              ScanRow[succ(succ(Adr))]:=Round(R);
            end;
     end;
   {zleva doprava}
   3:begin
      R:=RedX;G:=GreenX;B:=BlueX;
      ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
      For i:=succ(X[1]) to Y[1] do
      begin
       R:=R+StepR;G:=G+StepG;B:=B+StepB;
       Adr:=3*i;
       ScanRow[Adr]:=Round(B);
       ScanRow[succ(Adr)]:=Round(G);
       ScanRow[succ(succ(Adr))]:=Round(R);
      end;
    end;
   {k=1}
   4:begin
     R:=RedX;G:=GreenX;B:=BlueX;
     For i:=succ(X[1]) to Y[1] do
            begin
             inc(M[2]);
              R:=R+StepR;G:=G+StepG;B:=B+StepB;
              ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
              Adr:=3*i;
              ScanRow[Adr]:=Round(B);
              ScanRow[succ(Adr)]:=Round(G);
              ScanRow[succ(succ(Adr))]:=Round(R);
            end;
     end;
   {k=-1}
   5: begin
      R:=RedX;G:=GreenX;B:=BlueX;
      For i:=succ(X[1]) to Y[1] do
           begin
             dec(M[2]);
             R:=R+StepR;G:=G+StepG;B:=B+StepB;
             ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
             Adr:=3*i;
             ScanRow[Adr]:=Round(B);
             ScanRow[succ(Adr)]:=Round(G);
             ScanRow[succ(succ(Adr))]:=Round(R);
           end;
      end;
   {k>1}
   6:begin
       R:=RedX;G:=GreenX;B:=BlueX;
       p:=Dx;Dx:=Dy;Dy:=p;
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;
       While M[2]<Y[2] do
       begin
        Inc(M[2]);
        if p>0 then begin
                       Inc(M[1]);p:=p+p2;
                    end
                    else       p:=p+p1;
        ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
        Adr:=3*M[1];
        R:=R+StepR;G:=G+StepG;B:=B+StepB;
        ScanRow[Adr]:=Round(B);
        ScanRow[succ(Adr)]:=Round(G);
        ScanRow[succ(succ(Adr))]:=Round(R);
      end;
     end;
  {0<k<1}
  7: begin
       R:=RedX;G:=GreenX;B:=BlueX;
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;
       ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
       While M[1]<Y[1] do
       begin
        inc(M[1]);
        if p>0 then begin
                      inc(M[2]);p:=p+p2;
                      ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
                    end
                    else  p:=p+p1;
        Adr:=3*M[1];
        R:=R+StepR;G:=G+StepG;B:=B+StepB;
        ScanRow[Adr]:=Round(B);
        ScanRow[succ(Adr)]:=Round(G);
        ScanRow[succ(succ(Adr))]:=Round(R);
       end;
      end;
  {k<-1}
  8: begin
       R:=RedX;G:=GreenX;B:=BlueX;
       p:=Dx;Dx:=Dy;Dy:=-p;
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;
       While M[2]>Y[2] do
       begin
        dec(M[2]);
        if p<0 then begin
                     inc(M[1]);p:=p+p2;
                    end
                    else       p:=p+p1;
        ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
        Adr:=3*M[1];
        R:=R+StepR;G:=G+StepG;B:=B+StepB;
        ScanRow[Adr]:=Round(B);
        ScanRow[succ(Adr)]:=Round(G);
        ScanRow[succ(succ(Adr))]:=Round(R);
      end;
     end;
  {-1<k<0}
  9: begin
       R:=RedX;G:=GreenX;B:=BlueX;
       Dy:=-Dy;ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
       p1:=2*Dy;p2:=2*(Dy-Dx);p:=2*Dy-Dx;
       While M[1]<Y[1] do
       begin
        inc(M[1]);
        if p>0 then begin
                      dec(M[2]);
                      p:=p+p2;
                      ScanRow:=MemoryImage.Picture.Bitmap.ScanLine[M[2]];
                    end
                    else  p:=p+p1;
        Adr:=3*M[1];
        R:=R+StepR;G:=G+StepG;B:=B+StepB;
        ScanRow[Adr]:=Round(B);
        ScanRow[succ(Adr)]:=Round(G);
        ScanRow[succ(succ(Adr))]:=Round(R);
      end;
     end;
  end;
end;       {MemoryInterpolLine2D}

Procedure TDraw3D.Triangle2D(X,Y,Z:TPixel;Red,Green,Blue:Byte);
  begin
    Line2D(X,Y,Red,Green,Blue);
    Line2D(Y,Z,Red,Green,Blue);
    Line2D(X,Z,Red,Green,Blue);
  end;

Procedure TDraw3D.FillTriangle2D(X,Y,Z:TPixel;Red,Green,Blue:Byte);
  Const ProvisR=254;ProvisG=254;ProvisB=254;
  var i,j,h1,h2,MinI,MaxI,MinJ,MaxJ,Adr:Integer;
      BorderLine                       :Boolean;
       ScanRow: PByteArray;
  begin
    MinI:=X[1];MaxI:=X[1];MinJ:=X[2];MaxJ:=X[2];
    if MinI>Y[1] then MinI:=Y[1] else if MaxI<Y[1] then MaxI:=Y[1];
    if MinJ>Y[2] then MinJ:=Y[2] else if MaxJ<Y[2] then MaxJ:=Y[2];
    if MinI>Z[1] then MinI:=Z[1] else if MaxI<Z[1] then MaxI:=Z[1];
    if MinJ>Z[2] then MinJ:=Z[2] else if MaxJ<Z[2] then MaxJ:=Z[2];
    if (MinI<0) or (MaxI>=Image.Picture.Bitmap.Width) then exit;
    if (MinJ<0) or (MaxJ>=Image.Picture.Bitmap.Height) then exit;
    Line2D(X,Y,ProvisR,ProvisG,ProvisB);
    Line2D(Y,Z,ProvisR,ProvisG,ProvisB);
    Line2D(X,Z,ProvisR,ProvisG,ProvisB);
    For j:=MinJ to MaxJ do
    begin
       ScanRow:=Image.Picture.Bitmap.ScanLine[j];
       i:=Pred(MinI);
       Repeat
        inc(i);Adr:=3*i;
        BorderLine:=(ScanRow[Adr]=ProvisB) and (ScanRow[succ(Adr)]=ProvisG) and
                    (ScanRow[succ(succ(Adr))]=ProvisR)
       Until (i=MaxI) or BorderLine;
       if BorderLine
         then begin
               h1:=i;
               i:=succ(MaxI);
               Repeat
                 dec(i);Adr:=3*i;
                 BorderLine:=(ScanRow[Adr]=ProvisB) and (ScanRow[succ(Adr)]=ProvisG) and
                             (ScanRow[succ(succ(Adr))]=ProvisR)
               Until (i=h1) or BorderLine;
               h2:=i;
               For i:=h1 to h2 do
               begin
                 Adr:=3*i;
                 ScanRow[Adr]:=Blue;
                 ScanRow[succ(Adr)]:=Green;
                 ScanRow[succ(succ(Adr))]:=Red;
               end;
             end;
        end;
  end;         {FillTriangle2D}

Procedure TDraw3D.FillRoughTriangle2D(X,Y,Z:TPixel;Red,Green,Blue,Roughness:Byte;ColorDisp:Boolean);
  Const ProvisR=254;ProvisG=254;ProvisB=254;
  var i,j,h1,h2,MinI,MaxI,MinJ,MaxJ,Adr:Integer;
      BorderLine                       :Boolean;
      ScanRow                          :PByteArray;
      Rough,R,G,B                      :Byte;
      Factor                           :Double;
  begin
    Rough:=Trunc(Roughness*2.55);
    MinI:=X[1];MaxI:=X[1];MinJ:=X[2];MaxJ:=X[2];
    if MinI>Y[1] then MinI:=Y[1] else if MaxI<Y[1] then MaxI:=Y[1];
    if MinJ>Y[2] then MinJ:=Y[2] else if MaxJ<Y[2] then MaxJ:=Y[2];
    if MinI>Z[1] then MinI:=Z[1] else if MaxI<Z[1] then MaxI:=Z[1];
    if MinJ>Z[2] then MinJ:=Z[2] else if MaxJ<Z[2] then MaxJ:=Z[2];
    if (MinI<0) or (MaxI>=Image.Picture.Bitmap.Width) then exit;
    if (MinJ<0) or (MaxJ>=Image.Picture.Bitmap.Height) then exit;
    Line2D(X,Y,ProvisR,ProvisG,ProvisB);
    Line2D(Y,Z,ProvisR,ProvisG,ProvisB);
    Line2D(X,Z,ProvisR,ProvisG,ProvisB);
    For j:=MinJ to MaxJ do
    begin
       ScanRow:=Image.Picture.Bitmap.ScanLine[j];
       i:=Pred(MinI);
       Repeat
        inc(i);Adr:=3*i;
        BorderLine:=(ScanRow[Adr]=ProvisB) and (ScanRow[succ(Adr)]=ProvisG) and
                    (ScanRow[succ(succ(Adr))]=ProvisR)
       Until (i=MaxI) or BorderLine;
       if BorderLine
         then begin
               h1:=i;
               i:=succ(MaxI);
               Repeat
                 dec(i);Adr:=3*i;
                 BorderLine:=(ScanRow[Adr]=ProvisB) and (ScanRow[succ(Adr)]=ProvisG) and
                             (ScanRow[succ(succ(Adr))]=ProvisR)
               Until (i=h1) or BorderLine;
               h2:=i;
               For i:=h1 to h2 do
               begin
                 Factor:=(0.5-Random)*Rough;
                 if Red+Factor>255 then R:=255 else if Red+FActor<0 then R:=0 else R:=Trunc(Red+Factor);
                 if ColorDisp then  Factor:=(0.5-Random)*Rough;
                 if Green+Factor>255 then G:=255 else if Green+FActor<0 then G:=0 else G:=Trunc(Green+Factor);
                 if ColorDisp then  Factor:=(0.5-Random)*Rough;
                 if Blue+Factor>255 then B:=255 else if Blue+FActor<0 then B:=0 else B:=Trunc(Blue+Factor);
                 Adr:=3*i;
                 ScanRow[Adr]:=B;
                 ScanRow[succ(Adr)]:=G;
                 ScanRow[succ(succ(Adr))]:=R;
               end;
             end;
        end;
  end;         {FillRoughTriangle2D}

Procedure TDraw3D.FillTranspTriangle2D(X,Y,Z:TPixel;Red,Green,Blue:Byte;Transparency:Double);
 Const ProvisR=254;ProvisG=254;ProvisB=254;
 var i,j,MinI,MaxI,MinJ,MaxJ,Adr:Integer;
      BorderLine                       :Boolean;
      MemoryScanRow,ScanRow            :PByteArray;
      M,N                              :TPixel;
  begin
    MinI:=X[1];MaxI:=X[1];MinJ:=X[2];MaxJ:=X[2];
    if MinI>Y[1] then MinI:=Y[1] else if MaxI<Y[1] then MaxI:=Y[1];
    if MinJ>Y[2] then MinJ:=Y[2] else if MaxJ<Y[2] then MaxJ:=Y[2];
    if MinI>Z[1] then MinI:=Z[1] else if MaxI<Z[1] then MaxI:=Z[1];
    if MinJ>Z[2] then MinJ:=Z[2] else if MaxJ<Z[2] then MaxJ:=Z[2];
    if (MinI<0) or (MaxI>=MemoryImage.Picture.Bitmap.Width) then exit;
    if (MinJ<0) or (MaxJ>=MemoryImage.Picture.Bitmap.Height) then exit;
    MemoryLine2D(X,Y,ProvisR,ProvisG,ProvisB);
    MemoryLine2D(Y,Z,ProvisR,ProvisG,ProvisB);
    MemoryLine2D(X,Z,ProvisR,ProvisG,ProvisB);
    For j:=MinJ to MaxJ do
    begin
       MemoryScanRow:=MemoryImage.Picture.Bitmap.ScanLine[j];
             ScanRow:=      Image.Picture.Bitmap.ScanLine[j];
       i:=pred(MinI);
       Repeat
        inc(i);Adr:=3*i;
        BorderLine:=(MemoryScanRow[Adr]=ProvisB) and (MemoryScanRow[succ(Adr)]=ProvisG) and
                    (MemoryScanRow[succ(succ(Adr))]=ProvisR);
       Until (i=MaxI) or BorderLine;
       if BorderLine
         then begin
               M[1]:=i;M[2]:=j;
               i:=succ(MaxI);
               Repeat
                 dec(i);Adr:=3*i;
                 BorderLine:=(MemoryScanRow[Adr]=ProvisB) and (MemoryScanRow[succ(Adr)]=ProvisG) and
                        (MemoryScanRow[succ(succ(Adr))]=ProvisR);
               Until (i=M[1]) or BorderLine;
               N[1]:=i;N[2]:=j;
               For i:=M[1] to N[1] do
                 begin
                   Adr:=3*i;
                   ScanRow[Adr]:=Round((1-Transparency)*Blue+Transparency*ScanRow[Adr]);
                   ScanRow[succ(Adr)]:=Round((1-Transparency)*Green+Transparency*ScanRow[succ(Adr)]);
                   ScanRow[succ(succ(Adr))]:=Round((1-Transparency)*Red+Transparency*ScanRow[succ(succ(Adr))]);
                   MemoryScanRow[Adr]:=255;
                   MemoryScanRow[succ(Adr)]:=255;
                   MemoryScanRow[succ(succ(Adr))]:=255;
                 end;
             end;
       end;
 end;                  {FillTranspTriangle2D}

Procedure TDraw3D.FillInterpolTriangle2D(X,Y,Z:TPixel;RedX,GreenX,BlueX,
                                                      RedY,GreenY,BlueY,
                                                      RedZ,GreenZ,BlueZ:Byte);
 Const ProvisR=254;ProvisG=254;ProvisB=254;
  var i,j,MinI,MaxI,MinJ,MaxJ,Adr:Integer;
      BorderLine                       :Boolean;
      MemoryScanRow                    :PByteArray;
      R1,G1,B1,R2,G2,B2                :Byte;
      M,N                              :TPixel;
  begin

    MinI:=X[1];MaxI:=X[1];MinJ:=X[2];MaxJ:=X[2];
    if MinI>Y[1] then MinI:=Y[1] else if MaxI<Y[1] then MaxI:=Y[1];
    if MinJ>Y[2] then MinJ:=Y[2] else if MaxJ<Y[2] then MaxJ:=Y[2];
    if MinI>Z[1] then MinI:=Z[1] else if MaxI<Z[1] then MaxI:=Z[1];
    if MinJ>Z[2] then MinJ:=Z[2] else if MaxJ<Z[2] then MaxJ:=Z[2];
    if (MinI<0) or (MaxI>=MemoryImage.Picture.Bitmap.Width) then exit;
    if (MinJ<0) or (MaxJ>=MemoryImage.Picture.Bitmap.Height) then exit;
    MemoryLine2D(X,Y,ProvisR,ProvisG,ProvisB);
    MemoryLine2D(Y,Z,ProvisR,ProvisG,ProvisB);
    MemoryLine2D(X,Z,ProvisR,ProvisG,ProvisB);
    InterpolLine2D(X,Y,RedX,GreenX,BlueX,RedY,GreenY,BlueY);
    InterpolLine2D(Y,Z,RedY,GreenY,BlueY,RedZ,GreenZ,BlueZ);
    InterpolLine2D(X,Z,RedX,GreenX,BlueX,RedZ,GreenZ,BlueZ);
    For j:=MinJ to MaxJ do
    begin
       MemoryScanRow:=MemoryImage.Picture.Bitmap.ScanLine[j];
       i:=pred(MinI);
       Repeat
        inc(i);Adr:=3*i;
        BorderLine:=(MemoryScanRow[Adr]=ProvisB) and (MemoryScanRow[succ(Adr)]=ProvisG) and
                    (MemoryScanRow[succ(succ(Adr))]=ProvisR)
       Until (i=MaxI) or BorderLine;
       if BorderLine
         then begin
               M[1]:=i;M[2]:=j;
               i:=succ(MaxI);
               Repeat
                 dec(i);Adr:=3*i;
                 BorderLine:=(MemoryScanRow[Adr]=ProvisB) and (MemoryScanRow[succ(Adr)]=ProvisG) and
                             (MemoryScanRow[succ(succ(Adr))]=ProvisR)
               Until (i=M[1]) or BorderLine;
               N[1]:=i;N[2]:=J;
               GetPixel(M,R1,G1,B1);GetPixel(N,R2,G2,B2);
               InterpolLine2D(M,N,R1,G1,B1,R2,G2,B2);
             end;
    end;
    MemoryLine2D(X,Y,255,255,255);
    MemoryLine2D(Y,Z,255,255,255);
    MemoryLine2D(X,Z,255,255,255);
  end;                  {FillInterpolTriangle2D}

Procedure TDraw3D.FillInterpolTranspTriangle2D(X,Y,Z:TPixel;RedX,GreenX,BlueX,
                                                      RedY,GreenY,BlueY,
                                                      RedZ,GreenZ,BlueZ:Byte;Transparency:Double);
  var i,j,MinI,MaxI,MinJ,MaxJ,Adr:Integer;
      BorderLine                       :Boolean;
      MemoryScanRow,ScanRow            :PByteArray;
      R1,G1,B1,R2,G2,B2                :Byte;
      M,N                              :TPixel;
  begin
    MinI:=X[1];MaxI:=X[1];MinJ:=X[2];MaxJ:=X[2];
    if MinI>Y[1] then MinI:=Y[1] else if MaxI<Y[1] then MaxI:=Y[1];
    if MinJ>Y[2] then MinJ:=Y[2] else if MaxJ<Y[2] then MaxJ:=Y[2];
    if MinI>Z[1] then MinI:=Z[1] else if MaxI<Z[1] then MaxI:=Z[1];
    if MinJ>Z[2] then MinJ:=Z[2] else if MaxJ<Z[2] then MaxJ:=Z[2];
    if (MinI<0) or (MaxI>=MemoryImage.Picture.Bitmap.Width) then exit;
    if (MinJ<0) or (MaxJ>=MemoryImage.Picture.Bitmap.Height) then exit;
    MemoryInterpolLine2D(X,Y,RedX,GreenX,BlueX,RedY,GreenY,BlueY);
    MemoryInterpolLine2D(Y,Z,RedY,GreenY,BlueY,RedZ,GreenZ,BlueZ);
    MemoryInterpolLine2D(X,Z,RedX,GreenX,BlueX,RedZ,GreenZ,BlueZ);
    For j:=MinJ to MaxJ do
    begin
       MemoryScanRow:=MemoryImage.Picture.Bitmap.ScanLine[j];
             ScanRow:=      Image.Picture.Bitmap.ScanLine[j];
       i:=pred(MinI);
       Repeat
        inc(i);Adr:=3*i;
        BorderLine:=(MemoryScanRow[Adr]<>255) or (MemoryScanRow[succ(Adr)]<>255) or
                    (MemoryScanRow[succ(succ(Adr))]<>255)
       Until (i=MaxI) or BorderLine;
       if BorderLine
         then begin
               M[1]:=i;M[2]:=j;
               i:=succ(MaxI);
               Repeat
                 dec(i);Adr:=3*i;
                 BorderLine:=(MemoryScanRow[Adr]<>255) or (MemoryScanRow[succ(Adr)]<>255) or
                             (MemoryScanRow[succ(succ(Adr))]<>255)
               Until (i=M[1]) or BorderLine;
               N[1]:=i;N[2]:=j;
               Adr:=3*M[1];
               B1:=MemoryScanRow[Adr];
               G1:=MemoryScanRow[succ(Adr)];
               R1:=MemoryScanRow[succ(succ(Adr))];
               Adr:=3*N[1];
               B2:=MemoryScanRow[Adr];
               G2:=MemoryScanRow[succ(Adr)];
               R2:=MemoryScanRow[succ(succ(Adr))];
               MemoryInterpolLine2D(M,N,R1,G1,B1,R2,G2,B2);
               For i:=M[1] to N[1] do
                 begin
                   Adr:=3*i;
                   ScanRow[Adr]:=Round((1-Transparency)*MemoryScanRow[Adr]+Transparency*ScanRow[Adr]);
                   ScanRow[succ(Adr)]:=Round((1-Transparency)*MemoryScanRow[succ(Adr)]+Transparency*ScanRow[succ(Adr)]);
                   ScanRow[succ(succ(Adr))]:=Round((1-Transparency)*MemoryScanRow[succ(succ(Adr))]+Transparency*ScanRow[succ(succ(Adr))]);
                   MemoryScanRow[Adr]:=255;
                   MemoryScanRow[succ(Adr)]:=255;
                   MemoryScanRow[succ(succ(Adr))]:=255;
{                   if Pomalu then
                       begin
                          Image.Repaint;
                          Application.ProcessMessages;
                       end;}
                 end;
             end;
       end;
 end;                  {FillInterpolTranspTriangle2D}

procedure TDraw3D.Projection(X:T3DPoint;var Xp:T2DPoint);
  begin
    Xp[1]:=i1*X[1]+j1*X[2]+k1*X[3];
    Xp[2]:=i2*X[1]+j2*X[2]+k2*X[3];
  end;

procedure TDraw3D.SetOrthoAxonProjection(Alfa,Beta:Double);
  var u,v                :TVector;
      SinAlfa,i,j,
      CosGamma,CosDelta,PomAlfa,PomBeta:Double;
      O,S,X,Y,Z          :T3DPoint;

  begin
  ProjectionType:=ptOrtho;
  PomAlfa:=Alfa;PomBeta:=Beta;
  if PomAlfa>90 then PomAlfa:=PomAlfa-90;
  if PomAlfa>90 then PomAlfa:=PomAlfa-90;
  PomAlfa:=Pi*PomAlfa/180;PomBeta:=pi*PomBeta/180;
  S[1]:=cos(PomAlfa)*cos(PomBeta);S[2]:=sin(PomAlfa)*cos(PomBeta);S[3]:=sin(PomBeta);
  O[1]:=0;O[2]:=0;O[3]:=0;
  MajorRay[1]:=S[1];MajorRay[2]:=S[2];MajorRay[3]:=S[3];
  X[1]:=1/S[1];X[2]:=0;     X[3]:=0;
  Y[1]:=0;     Y[2]:=1/S[2];Y[3]:=0;
  Z[1]:=0;     Z[2]:=0;     Z[3]:=1/S[3];
  SetVector(S,X,u);SetVector(S,Z,v);
  CosGamma:=ScalarProduct(u,v)/Norm(u)/Norm(v);
  SetVector(S,Y,u);SetVector(S,Z,v);
  CosDelta:=ScalarProduct(u,v)/Norm(u)/Norm(v);
  SetVector(X,O,u);SetVector(X,S,v);
  i:=ScalarProduct(u,v)/Norm(u)/Norm(v);
  SetVector(Y,O,u);SetVector(Y,S,v);
  j:=ScalarProduct(u,v)/Norm(u)/Norm(v);
  SetVector(Z,O,u);SetVector(Z,S,v);
  i1:=-i*sqrt(1-CosGamma*CosGamma);i2:=i*CosGamma;
  j1:=j*sqrt(1-CosDelta*CosDelta);j2:=j*CosDelta;
  k1:=0;                          k2:=cos(PomBeta);
  if Alfa>90 then
  begin
  S[1]:=cos(Alfa*pi/180)*cos(Beta*pi/180);S[2]:=sin(Alfa*pi/180)*cos(Beta*pi/180);S[3]:=sin(Beta*pi/180);
  MajorRay[1]:=S[1];MajorRay[2]:=S[2];MajorRay[3]:=S[3];
  if Alfa<270 then
  Repeat
    Alfa:=Alfa-90;
    begin
    sinAlfa:=i1;i1:= j1;j1:=SinAlfa;
    sinAlfa:=i2;i2:=j2;j2:=SinAlfa;
    i1:=-i1;i2:=-i2;
    end;
   Until Alfa<90
   else begin
         i1:= -i1;j2:=-j2;
        end;
  end;
end;

Procedure TDraw3D.SetCenterProjection(Alfa,Beta,ObserverDistance:Double;var ObserverPoint:T3DPoint);
var u,v                :TVector;
      SinAlfa,i,j,
      CosGamma,CosDelta,PomAlfa,PomBeta:Double;
      O,S,X,Y,Z          :T3DPoint;
begin
  ProjectionType:=ptCenter;
  ObserverPoint[1]:=ObserverDistance*cos(Alfa*pi/180)*cos(Beta*pi/180);
  ObserverPoint[2]:=ObserverDistance*sin(Alfa*pi/180)*cos(Beta*pi/180);
  ObserverPoint[3]:=ObserverDistance*sin(Beta*pi/180);
  PomAlfa:=Alfa;PomBeta:=Beta;
  if PomAlfa>90 then PomAlfa:=PomAlfa-90;
  if PomAlfa>90 then PomAlfa:=PomAlfa-90;
  PomAlfa:=Pi*PomAlfa/180;PomBeta:=pi*PomBeta/180;
  S[1]:=cos(PomAlfa)*cos(PomBeta);S[2]:=sin(PomAlfa)*cos(PomBeta);S[3]:=sin(PomBeta);
  O[1]:=0;O[2]:=0;O[3]:=0;
  MajorRay[1]:=S[1];MajorRay[2]:=S[2];MajorRay[3]:=S[3];
  X[1]:=1/S[1];X[2]:=0;     X[3]:=0;
  Y[1]:=0;     Y[2]:=1/S[2];Y[3]:=0;
  Z[1]:=0;     Z[2]:=0;     Z[3]:=1/S[3];
  SetVector(S,X,u);SetVector(S,Z,v);
  CosGamma:=ScalarProduct(u,v)/Norm(u)/Norm(v);
  SetVector(S,Y,u);SetVector(S,Z,v);
  CosDelta:=ScalarProduct(u,v)/Norm(u)/Norm(v);
  SetVector(X,O,u);SetVector(X,S,v);
  i:=ScalarProduct(u,v)/Norm(u)/Norm(v);
  SetVector(Y,O,u);SetVector(Y,S,v);
  j:=ScalarProduct(u,v)/Norm(u)/Norm(v);
  SetVector(Z,O,u);SetVector(Z,S,v);
  i1:=-i*sqrt(1-CosGamma*CosGamma);i2:=i*CosGamma;
  j1:=j*sqrt(1-CosDelta*CosDelta);j2:=j*CosDelta;
  k1:=0;                          k2:=cos(PomBeta);
  if Alfa>90 then
  begin
  S[1]:=cos(Alfa*pi/180)*cos(Beta*pi/180);S[2]:=sin(Alfa*pi/180)*cos(Beta*pi/180);S[3]:=sin(Beta*pi/180);
  MajorRay[1]:=S[1];MajorRay[2]:=S[2];MajorRay[3]:=S[3];
  if Alfa<270 then
  Repeat
    Alfa:=Alfa-90;
    begin
    sinAlfa:=i1;i1:= j1;j1:=SinAlfa;
    sinAlfa:=i2;i2:=j2;j2:=SinAlfa;
    i1:=-i1;i2:=-i2;
    end;
   Until Alfa<90
   else begin
         i1:= -i1;j2:=-j2;
        end;
  end;
end;

Procedure TDraw3D.SetSphereProjection(Alfa,Beta,ObserverDistance:Double;var ObserverPoint:T3DPoint);
var u,v                :TVector;
      SinAlfa,i,j,
      CosGamma,CosDelta,PomAlfa,PomBeta:Double;
      O,S,X,Y,Z          :T3DPoint;
begin
  ProjectionType:=ptSphere;
  ObserverPoint[1]:=ObserverDistance*cos(Alfa*pi/180)*cos(Beta*pi/180);
  ObserverPoint[2]:=ObserverDistance*sin(Alfa*pi/180)*cos(Beta*pi/180);
  ObserverPoint[3]:=ObserverDistance*sin(Beta*pi/180);
  PomAlfa:=Alfa;PomBeta:=Beta;
  if PomAlfa>90 then PomAlfa:=PomAlfa-90;
  if PomAlfa>90 then PomAlfa:=PomAlfa-90;
  PomAlfa:=Pi*PomAlfa/180;PomBeta:=pi*PomBeta/180;
  S[1]:=cos(PomAlfa)*cos(PomBeta);S[2]:=sin(PomAlfa)*cos(PomBeta);S[3]:=sin(PomBeta);
  O[1]:=0;O[2]:=0;O[3]:=0;
  MajorRay[1]:=S[1];MajorRay[2]:=S[2];MajorRay[3]:=S[3];
  X[1]:=1/S[1];X[2]:=0;     X[3]:=0;
  Y[1]:=0;     Y[2]:=1/S[2];Y[3]:=0;
  Z[1]:=0;     Z[2]:=0;     Z[3]:=1/S[3];
  SetVector(S,X,u);SetVector(S,Z,v);
  CosGamma:=ScalarProduct(u,v)/Norm(u)/Norm(v);
  SetVector(S,Y,u);SetVector(S,Z,v);
  CosDelta:=ScalarProduct(u,v)/Norm(u)/Norm(v);
  SetVector(X,O,u);SetVector(X,S,v);
  i:=ScalarProduct(u,v)/Norm(u)/Norm(v);
  SetVector(Y,O,u);SetVector(Y,S,v);
  j:=ScalarProduct(u,v)/Norm(u)/Norm(v);
  SetVector(Z,O,u);SetVector(Z,S,v);
  i1:=-i*sqrt(1-CosGamma*CosGamma);i2:=i*CosGamma;
  j1:=j*sqrt(1-CosDelta*CosDelta);j2:=j*CosDelta;
  k1:=0;                          k2:=cos(PomBeta);
  if Alfa>90 then
  begin
  S[1]:=cos(Alfa*pi/180)*cos(Beta*pi/180);S[2]:=sin(Alfa*pi/180)*cos(Beta*pi/180);S[3]:=sin(Beta*pi/180);
  MajorRay[1]:=S[1];MajorRay[2]:=S[2];MajorRay[3]:=S[3];
  if Alfa<270 then
  Repeat
    Alfa:=Alfa-90;
    begin
    sinAlfa:=i1;i1:= j1;j1:=SinAlfa;
    sinAlfa:=i2;i2:=j2;j2:=SinAlfa;
    i1:=-i1;i2:=-i2;
    end;
   Until Alfa<90
   else begin
         i1:= -i1;j2:=-j2;
        end;
  end;
end;

procedure TDraw3D.SetDirectOfLight(AlfaLight,BetaLight:Double;var DirectOfLight:TVector);
begin
  DirectOfLight[1]:=cos(pi*AlfaLight/180)*cos(pi*BetaLight/180);
  DirectOfLight[2]:=sin(pi*AlfaLight/180)*cos(pi*BetaLight/180);
  DirectOfLight[3]:=sin(pi*BetaLight/180);
end;

 Procedure TDraw3D.SetVector(X,Y:T3DPoint;var v:TVector);
  begin
    v[1]:=Y[1]-X[1];
    v[2]:=Y[2]-X[2];
    v[3]:=Y[3]-X[3];
  end;

  Function TDraw3D.ScalarProduct(u,v:TVector):Double;
  begin
    ScalarProduct:=u[1]*v[1]+u[2]*v[2]+u[3]*v[3];
  end;

  Procedure TDraw3D.VectorProduct(r,s:TVector;var t:TVector);
  begin
   t[1]:= r[2]*s[3]-s[2]*r[3];
   t[2]:=-r[1]*s[3]+s[1]*r[3];
   t[3]:= r[1]*s[2]-s[1]*r[2];
  end;

  Function TDraw3D.Norm(u:TVector):Double;
  begin
    Norm:=Sqrt(u[1]*u[1]+u[2]*u[2]+u[3]*u[3]);
  end;

  function TDraw3D.CosAngle(r,s:TVector):Double;
  begin
    if (Norm(r)<1e-25) or (Norm(s)<1e-25)
     then CosAngle:=0
     else CosAngle:=ScalarProduct(r,s)/Norm(r)/Norm(s);
  end;

Procedure TDraw3D.PlaneIntersect(X:T3DPoint;var pX:T3DPoint);
var t:Double;
    Sm:TVector;
begin
  Sm:=MajorRay;
  t:=-Sm[1]*ObserverPoint[1]-Sm[2]*ObserverPoint[2]-Sm[3]*ObserverPoint[3];
  t:=t/(Sm[1]*(X[1]-ObserverPoint[1])+Sm[2]*(X[2]-ObserverPoint[2])+Sm[3]*(X[3]-ObserverPoint[3]));
  pX[1]:=ObserverPoint[1]+(X[1]-ObserverPoint[1])*t;
  pX[2]:=ObserverPoint[2]+(X[2]-ObserverPoint[2])*t;
  pX[3]:=ObserverPoint[3]+(X[3]-ObserverPoint[3])*t;
end;

Procedure TDraw3D.SphereIntersect(X:T3DPoint;var pX:T3DPoint);
var t:Double;
begin
  t:=sqrt(sqr(ObserverPoint[1])+sqr(ObserverPoint[2])+sqr(ObserverPoint[3]))-1;
  t:=t/sqrt(sqr(ObserverPoint[1]-x[1])+sqr(ObserverPoint[2]-x[2])+sqr(ObserverPoint[3]-x[3]));
  pX[1]:=ObserverPoint[1]+(X[1]-ObserverPoint[1])*t;
  pX[2]:=ObserverPoint[2]+(X[2]-ObserverPoint[2])*t;
  pX[3]:=ObserverPoint[3]+(X[3]-ObserverPoint[3])*t;
end;

Procedure  TDraw3D.Line(A,B:T3DPoint;Red,Green,Blue:Byte);
var  IntSectA,IntSectB:T3DPoint;
     pX,pY            :T2DPoint;
     X,Y              :TPixel;
     t                :Double;
     i                :Integer;
 begin
   Case ProjectionType of
       {axonometrie}
       0,1: begin IntSectA:=A;IntSectB:=B;end;
       {stedov promtn na rovinu}
       2: begin PlaneIntersect(A,IntSectA);PlaneIntersect(B,IntSectB);end;
    end;
    if ProjectionType<3
         then begin
              Projection(IntSectA,pX);Projection(IntSectB,pY);
              X[1]:=XCoor(pX[1]);X[2]:=YCoor(pX[2]);
              Y[1]:=XCoor(pY[1]);Y[2]:=YCoor(pY[2]);
              Line2D(X,Y,Red,Green,Blue);
              end
         else begin  {stedov promtn na kulovou plochu}
               t:=0;i:=0;
               While t<=1.05 do
                 begin
                  P[i,1]:=A[1]+(B[1]-A[1])*t;
                  P[i,2]:=A[2]+(B[2]-A[2])*t;
                  P[i,3]:=A[3]+(B[3]-A[3])*t;
                  SphereIntersect(P[i],Q[i]);
                  i:=succ(i);t:=t+0.05;
                 end;
               PolyLine(Q,i,Red,Green,Blue)
              end;
 end;

procedure TDraw3D.PolyLine(X:TArrayOf3DPoints;n:Word;Red,Green,Blue:Byte);
 var i:Word;
 begin
   for i:=0 to n-2 do
     begin
      Line(X[i],X[succ(i)],Red,Green,Blue);
 {     Image.Repaint;}
     end; 
 end;

 Procedure TDraw3D.Put3DPoint(X:T3DPoint;Red,Green,Blue:Byte);
 var IntSectX                :T3DPoint;
     pX                      :T2DPoint;
     PixelX                  :TPixel;
 begin
   Case ProjectionType of
       {axonometrie}
       0,1:IntSectX:=X;
       {stedov promtn na rovinu}
       2:PlaneIntersect(X,IntSectX);
       {stedov promtn na kulovou plochu}
       3:SphereIntersect(X,IntSectX);
    end;
    Projection(IntSectX,pX);
    PixelX[1]:=XCoor(pX[1]);PixelX[2]:=YCoor(pX[2]);
    PutPixel(PixelX,Red,Green,Blue);
 end;   

 Procedure TDraw3D.Cross(X:T3DPoint;Size,Red,Green,Blue:Byte);
 var IntSectX                :T3DPoint;
     pX                      :T2DPoint;
     PixelX,A,B,C,D          :TPixel;
 begin
   Case ProjectionType of
       {axonometrie}
       0,1:IntSectX:=X;
       {stedov promtn na rovinu}
       2:PlaneIntersect(X,IntSectX);
       {stedov promtn na kulovou plochu}
       3:SphereIntersect(X,IntSectX);
    end;
    Projection(IntSectX,pX);
    PixelX[1]:=XCoor(pX[1]);PixelX[2]:=YCoor(pX[2]);
    A[1]:=PixelX[1]-Size;A[2]:=PixelX[2];
    B[1]:=PixelX[1]+Size;B[2]:=PixelX[2];
    C[1]:=PixelX[1];     C[2]:=PixelX[2]-Size;
    D[1]:=PixelX[1];     D[2]:=PixelX[2]+Size;
    Line2D(A,B,Red,Green,Blue);
    Line2D(C,D,Red,Green,Blue);
end;

{procedure TDraw3D.Line(X,Y:T3DPoint;Red,Green,Blue:Byte);
var A,B  :TPixel;
  begin
     A[1]:=XCoor(X[1]);A[2]:=YCoor(X[2]);
     B[1]:=XCoor(Y[1]);B[2]:=YCoor(Y[2]);
     Line2D(A,B,Red,Green,Blue);
  end;}

 {Sestroj trojhelnk vyplnn zadanou barvou}
 procedure TDraw3D.FillTriangle(A,B,C:T3DPoint;Red,Green,Blue:byte);
 var IntSectA,IntSectB,IntSectC:T3DPoint;
     pX,pY,pZ                :T2DPoint;
     X,Y,Z                   :TPixel;
 begin
   Case ProjectionType of
       {axonometrie}
       0,1: begin
           IntSectA:=A;IntSectB:=B;IntSectC:=C;
          end;
       {stedov promtn na rovinu}
       2: begin
            PlaneIntersect(A,IntSectA);PlaneIntersect(B,IntSectB);PlaneIntersect(C,IntSectC);
          end;
       {stedov promtn na kulovou plochu}
       3: begin
            SphereIntersect(A,IntSectA);SphereIntersect(B,IntSectB);SphereIntersect(C,IntSectC);
          end;
    end;
    Projection(IntSectA,pX);Projection(IntSectB,pY);Projection(IntSectC,pZ);
    X[1]:=XCoor(pX[1]);X[2]:=YCoor(pX[2]);
    Y[1]:=XCoor(pY[1]);Y[2]:=YCoor(pY[2]);
    Z[1]:=XCoor(pZ[1]);Z[2]:=YCoor(pZ[2]);
    FillTriangle2D(X,Y,Z,Red,Green,Blue);
 end;

{Sestroj drsn trojhelnk vyplnn zadanou barvou}
 procedure TDraw3D.FillRoughTriangle(A,B,C:T3DPoint;Red,Green,Blue,Roughness:byte;ColorDisp:Boolean);
 var IntSectA,IntSectB,IntSectC:T3DPoint;
     pX,pY,pZ                :T2DPoint;
     X,Y,Z                   :TPixel;
 begin
   Case ProjectionType of
       {axonometrie}
       0,1: begin
           IntSectA:=A;IntSectB:=B;IntSectC:=C;
          end;
       {stedov promtn na rovinu}
       2: begin
            PlaneIntersect(A,IntSectA);PlaneIntersect(B,IntSectB);PlaneIntersect(C,IntSectC);
          end;
       {stedov promtn na kulovou plochu}
       3: begin
            SphereIntersect(A,IntSectA);SphereIntersect(B,IntSectB);SphereIntersect(C,IntSectC);
          end;
    end;
    Projection(IntSectA,pX);Projection(IntSectB,pY);Projection(IntSectC,pZ);
    X[1]:=XCoor(pX[1]);X[2]:=YCoor(pX[2]);
    Y[1]:=XCoor(pY[1]);Y[2]:=YCoor(pY[2]);
    Z[1]:=XCoor(pZ[1]);Z[2]:=YCoor(pZ[2]);
    FillRoughTriangle2D(X,Y,Z,Red,Green,Blue,Roughness,ColorDisp);
 end;

    {Sestroj prhledn trojhelnk vyplnn zadanou barvou}
 procedure TDraw3D.FillTranspTriangle(A,B,C:T3DPoint;Red,Green,Blue:byte;Transparency:Double);
 var IntSectA,IntSectB,IntSectC:T3DPoint;
     pX,pY,pZ                :T2DPoint;
     X,Y,Z                   :TPixel;
 begin
   Case ProjectionType of
       {axonometrie}
       0,1: begin
           IntSectA:=A;IntSectB:=B;IntSectC:=C;
          end;
       {stedov promtn na rovinu}
       2: begin
            PlaneIntersect(A,IntSectA);PlaneIntersect(B,IntSectB);PlaneIntersect(C,IntSectC);
          end;
       {stedov promtn na kulovou plochu}
       3: begin
            SphereIntersect(A,IntSectA);SphereIntersect(B,IntSectB);SphereIntersect(C,IntSectC);
          end;
    end;
    Projection(IntSectA,pX);Projection(IntSectB,pY);Projection(IntSectC,pZ);
    X[1]:=XCoor(pX[1]);X[2]:=YCoor(pX[2]);
    Y[1]:=XCoor(pY[1]);Y[2]:=YCoor(pY[2]);
    Z[1]:=XCoor(pZ[1]);Z[2]:=YCoor(pZ[2]);
    FillTranspTriangle2D(X,Y,Z,Red,Green,Blue,Transparency);
 end;

 {Sestroj trojhelnk vyplnn interpolac barvy ve vrcholech}
 procedure TDraw3D.FillInterpolTriangle(A,B,C:T3DPoint;RedA,GreenA,BlueA,
                                                    RedB,GreenB,BlueB,
                                                    RedC,GreenC,BlueC:byte);
 var IntSectA,IntSectB,IntSectC:T3DPoint;
     pX,pY,pZ                :T2DPoint;
     X,Y,Z                   :TPixel;
 begin
   Case ProjectionType of
       {axonometrie}
       0,1: begin
           IntSectA:=A;IntSectB:=B;IntSectC:=C;
          end;
       {stedov promtn na rovinu}
       2: begin
            PlaneIntersect(A,IntSectA);PlaneIntersect(B,IntSectB);PlaneIntersect(C,IntSectC);
          end;
       {stedov promtn na kulovou plochu}
       3: begin
            SphereIntersect(A,IntSectA);SphereIntersect(B,IntSectB);SphereIntersect(C,IntSectC);
          end;
    end;
    Projection(IntSectA,pX);Projection(IntSectB,pY);Projection(IntSectC,pZ);
    X[1]:=XCoor(pX[1]);X[2]:=YCoor(pX[2]);
    Y[1]:=XCoor(pY[1]);Y[2]:=YCoor(pY[2]);
    Z[1]:=XCoor(pZ[1]);Z[2]:=YCoor(pZ[2]);
    FillInterpolTriangle2D(X,Y,Z,RedA,GreenA,BlueA,RedB,GreenB,BlueB,RedC,GreenC,BlueC);
 end;

 {Sestroj prhledn trojhelnk vyplnn interpolac barvy ve vrcholech}
 procedure TDraw3D.FillInterpolTranspTriangle
                        (A,B,C:T3DPoint;RedA,GreenA,BlueA,
                                        RedB,GreenB,BlueB,
                                        RedC,GreenC,BlueC:byte;Transparency:Double);
 var IntSectA,IntSectB,IntSectC:T3DPoint;
     pX,pY,pZ                :T2DPoint;
     X,Y,Z                   :TPixel;
 begin
   Case ProjectionType of
       {axonometrie}
       0,1: begin
           IntSectA:=A;IntSectB:=B;IntSectC:=C;
          end;
       {stedov promtn na rovinu}
       2: begin
            PlaneIntersect(A,IntSectA);PlaneIntersect(B,IntSectB);PlaneIntersect(C,IntSectC);
          end;
       {stedov promtn na kulovou plochu}
       3: begin
            SphereIntersect(A,IntSectA);SphereIntersect(B,IntSectB);SphereIntersect(C,IntSectC);
          end;
    end;
    Projection(IntSectA,pX);Projection(IntSectB,pY);Projection(IntSectC,pZ);
    X[1]:=XCoor(pX[1]);X[2]:=YCoor(pX[2]);
    Y[1]:=XCoor(pY[1]);Y[2]:=YCoor(pY[2]);
    Z[1]:=XCoor(pZ[1]);Z[2]:=YCoor(pZ[2]);
    FillInterpolTranspTriangle2D(X,Y,Z,RedA,GreenA,BlueA,RedB,GreenB,BlueB,
                                       RedC,GreenC,BlueC,Transparency);
 end;

(* Procedure TDraw3D.XAxis(x1,x2,y:Double;Red,Green,Blue:Byte);
 begin
    A[1]:=x1;A[2]:=y;B[1]:=x2;B[2]:=A[2];
    cvLine(A,B,Red,Green,Blue);
 end;

Procedure TDraw3D.YAxis(y1,y2,x:Double;Red,Green,Blue:Byte);
  begin
    A[1]:=x;A[2]:=y1;B[1]:=A[1];B[2]:=y2;
    cvLine(A,B,Red,Green,Blue);
  end;

procedure TDraw3D.GaugedXAxis(x1,x2,y:Double;Red,Green,Blue:Byte);
  var  AxisUnit        :Double;
       Text            :String;
       Lx,Exponent     :Integer;

   begin
   XAxis(x1,x2,y,Red,Green,Blue);
   Image.Canvas.Font.Color:=Red+256*Green+256*256*Blue;
   Image.Canvas.Pen.Color:=Red+256*Green+256*256*Blue;
   AxisUnit:=1E30;Lx:=0;Exponent:=30;
   While (x2-x1)/AxisUnit<3 do
    begin
       AxisUnit:=AxisUnit/10;Exponent:=Exponent-1;
    end;
   Lx:=0;
   if x1<0 then
     Repeat Lx:=Lx-1 Until AxisUnit*(Lx-1)<x1
           else
     Repeat Lx:=Lx+1 Until AxisUnit*(Lx+1)>x1;
   With Image.Canvas do
   repeat
      MoveTo(XCoor(AxisUnit*Lx),YCoor(y)-3);
      LineTo(XCoor(AxisUnit*Lx),YCoor(y)+3);
      Str(Lx,Text);
      TextOut(XCoor(AxisUnit*Lx)+3,YCoor(y)+5,Text);
      Lx:=Lx+1;
   until AxisUnit*(Lx+1)>x2;
   Str(Exponent,Text);
   Image.Canvas.Brush.Color:=clWhite;
  { Image.Canvas.TextOut(XCoor(x2)-40,YCoor(y)+5,'    ');}
   Image.Canvas.TextOut(XCoor(x2)-40,YCoor(y)+5,'*10');
   Image.Canvas.TextOut(XCoor(x2)-20,YCoor(Y)+1,Text);
   end;

procedure TDraw3D.GaugedYAxis(y1,y2,x:Double;Red,Green,Blue:Byte);
   var AxisUnit        :Double;
       Text            :String;
       Ly,Exponent     :Integer;
   begin
   YAxis(y1,y2,x,Red,Green,Blue);
   Image.Canvas.Font.Color:=Red+256*Green+256*256*Blue;
   Image.Canvas.Pen.Color:=Red+256*Green+256*256*Blue;
   AxisUnit:=1E30;Ly:=0;Exponent:=30;
   While (y2-y1)/AxisUnit<3 do
    begin
       AxisUnit:=AxisUnit/10;Exponent:=Exponent-1;
    end;
   Ly:=0;
   if y1<0 then
     Repeat Ly:=Ly-1 Until AxisUnit*(Ly-1)<y1
           else
     Repeat Ly:=Ly+1 Until AxisUnit*(Ly+1)>y1;
   With Image.Canvas do
   repeat
      MoveTo(XCoor(x)-3,YCoor(AxisUnit*Ly));
      LineTo(XCoor(x)+3,YCoor(AxisUnit*Ly));
      Str(Ly,Text);
      TextOut(XCoor(x)-25,YCoor(AxisUnit*Ly)-15,Text);
      Ly:=Ly+1;
   until AxisUnit*(Ly+1)>y2;
   Str(Exponent,Text);
   Image.Canvas.TextOut(XCoor(x)+5,YCoor(y2)+10,'    ');
   Image.Canvas.TextOut(XCoor(x)+5,YCoor(y2)+10,'*10');
   Image.Canvas.TextOut(XCoor(x)+25,YCoor(y2)+3,Text);
   end; *)

 Procedure TDraw3D.Triangle(A,B,C:T3DPoint;Red,Green,Blue:Byte);
 begin
   Line(A,B,Red,Green,Blue);
   Line(B,C,Red,Green,Blue);
   Line(C,A,Red,Green,Blue);
 end;

procedure TDraw3D.cvLine(A,B:T3DPoint;Red,Green,Blue:byte);
var  IntSectA,IntSectB:T3DPoint;
     pX,pY            :T2DPoint;
     X,Y              :TPixel;
     t                :Double;
     i,j              :Integer;
 begin
   Case ProjectionType of
       {axonometrie}
       0,1: begin IntSectA:=A;IntSectB:=B;end;
       {stedov promtn na rovinu}
       2: begin PlaneIntersect(A,IntSectA);PlaneIntersect(B,IntSectB);end;
    end;
      if ProjectionType<3
         then begin
              Projection(IntSectA,pX);Projection(IntSectB,pY);
              X[1]:=XCoor(pX[1]);X[2]:=YCoor(pX[2]);
              Y[1]:=XCoor(pY[1]);Y[2]:=YCoor(pY[2]);
              With Image.Canvas do
                begin
                  Pen.Color:=Red+256*Green+256*256*Blue;
                  MoveTo(X[1],X[2]);
                  LineTo(Y[1],Y[2]);
                end;
              end
         else begin  {stedov promtn na kulovou plochu}
               t:=0;i:=0;
               While t<=1.05 do
                 begin
                  P[i,1]:=A[1]+(B[1]-A[1])*t;
                  P[i,2]:=A[2]+(B[2]-A[2])*t;
                  P[i,3]:=A[3]+(B[3]-A[3])*t;
                  SphereIntersect(P[i],Q[i]);
                  i:=succ(i);t:=t+0.05;
                 end;
               Projection(Q[0],pX);Image.Canvas.Pen.Color:=Red+256*Green+256*256*Blue;
               Image.Canvas.MoveTo(XCoor(pX[1]),YCoor(pX[2]));
               for i:=1 to i-1 do
                  begin
                    Projection(Q[i],pX);
                    Image.Canvas.LineTo(XCoor(pX[1]),YCoor(pX[2]));
                  end;
              end;
end;

procedure TDraw3D.cvPolyLine(X:TArrayOf3DPoints;n:Word;Red,Green,Blue:byte);
var i:Word;
begin
  {  With Image.Canvas do
     begin
       Pen.Color:=Red+256*Green+256*256*Blue;
       MoveTo(XCoor(X[0,1]),YCoor(X[0,2]));
       for i:=1 to n-1 do
         LineTo(XCoor(X[i,1]),YCoor(X[i,2]));
    end;}
    for i:=0 to n-2 do cvLine(X[i],X[succ(i)],Red,Green,Blue);
end;

procedure TDraw3D.cvBox(Box:TBox;Red,Green,Blue:byte);
begin
 With Box do
 begin
  cvLine(A,B,Red,Green,Blue);cvLine(B,C,Red,Green,Blue);
  cvLine(C,D,Red,Green,Blue);cvLine(D,A,Red,Green,Blue);
  cvLine(Ac,Bc,Red,Green,Blue);cvLine(Bc,Cc,Red,Green,Blue);
  cvLine(Cc,Dc,Red,Green,Blue);cvLine(Dc,Ac,Red,Green,Blue);
  cvLine(A,Ac,Red,Green,Blue);cvLine(B,Bc,Red,Green,Blue);
  cvLine(C,Cc,Red,Green,Blue);cvLine(D,Dc,Red,Green,Blue);
 end;
end;

procedure TDraw3D.TransOfPoint(X:T3DPoint;Z:TMatrix;var Xc:T3DPoint);
   var i,j:Byte;
   begin
      for i:=1 to 4 do Xc[i]:=0;
      for j:=1 to 4 do
       for i:=1 to 4 do
           Xc[j]:=Xc[j]+X[i]*Z[i,j];
   end;

   Procedure TDraw3D.TransOfTriangle(A,B,C:T3DPoint;Z:TMatrix;
                 var Ac,Bc,Cc:T3DPoint);
   begin
      TransOfPoint(A,Z,Ac);
      TransOfPoint(B,Z,Bc);
      TransOfPoint(C,Z,Cc);
   end;

Procedure TDraw3D.TransOfBox(Box1:TBox;Z:TMatrix;var Box2:TBox);
   begin
      TransOfPoint(Box1.A,Z,Box2.A);
      TransOfPoint(Box1.B,Z,Box2.B);
      TransOfPoint(Box1.C,Z,Box2.C);
      TransOfPoint(Box1.D,Z,Box2.D);
      TransOfPoint(Box1.Ac,Z,Box2.Ac);
      TransOfPoint(Box1.Bc,Z,Box2.Bc);
      TransOfPoint(Box1.Cc,Z,Box2.Cc);
      TransOfPoint(Box1.Dc,Z,Box2.Dc);
   end;

Procedure TDraw3D.Multip(M,N:TMatrix;var Result:TMatrix);
Var i,j,k:Integer;
begin
    for i:=1 to 4 do for j:=1 to 4 do Result[i,j]:=0;
    for i:=1 to 4 do for k:=1 to 4 do for j:=1 to 4 do
                      Result[i,j]:=Result[i,j]+M[i,k]*N[k,j];
end;

procedure TDraw3D.Transl(v:TVector;var M:TMatrix);
begin
  M[1,1]:=1;   M[1,2]:=0;   M[1,3]:=0;   M[1,4]:=0;
  M[2,1]:=0;   M[2,2]:=1;   M[2,3]:=0;   M[2,4]:=0;
  M[3,1]:=0;   M[3,2]:=0;   M[3,3]:=1;   M[3,4]:=0;
  M[4,1]:=v[1];M[4,2]:=v[2];M[4,3]:=v[3];M[4,4]:=1;
end;

procedure TDraw3D.RotX(Alfa:Double;var M:TMatrix);
begin
  M[1,1]:=1;M[1,2]:= 0;        M[1,3]:=0;        M[1,4]:=0;
  M[2,1]:=0;M[2,2]:= cos(Alfa);M[2,3]:=sin(Alfa);M[2,4]:=0;
  M[3,1]:=0;M[3,2]:=-sin(alfa);M[3,3]:=cos(Alfa);M[3,4]:=0;
  M[4,1]:=0;M[4,2]:= 0;        M[4,3]:=0;        M[4,4]:=1;
end;

procedure TDraw3D.RotY(Beta:Double;var M:TMatrix);
begin
  M[1,1]:= cos(Beta);M[1,2]:=0;M[1,3]:=sin(Beta);M[1,4]:=0;
  M[2,1]:= 0;        M[2,2]:=1;M[2,3]:=0;        M[2,4]:=0;
  M[3,1]:=-sin(Beta);M[3,2]:=0;M[3,3]:=cos(Beta);M[3,4]:=0;
  M[4,1]:= 0;        M[4,2]:=0;M[4,3]:=0;        M[4,4]:=1;
end;

procedure TDraw3D.RotZ(Gamma:Double;var M:TMatrix);
begin
  M[1,1]:= cos(Gamma);M[1,2]:=sin(Gamma);M[1,3]:=0;M[1,4]:=0;
  M[2,1]:=-sin(Gamma);M[2,2]:=cos(Gamma);M[2,3]:=0;M[2,4]:=0;
  M[3,1]:= 0;         M[3,2]:=0;         M[3,3]:=1;M[3,4]:=0;
  M[4,1]:= 0;         M[4,2]:=0;         M[4,3]:=0;M[4,4]:=1;
end;

procedure TDraw3D.AxisSymX(var M:TMatrix);
begin
  M[1,1]:=1;M[1,2]:= 0;M[1,3]:= 0;M[1,4]:=0;
  M[2,1]:=0;M[2,2]:=-1;M[2,3]:= 0;M[2,4]:=0;
  M[3,1]:=0;M[3,2]:= 0;M[3,3]:=-1;M[3,4]:=0;
  M[4,1]:=0;M[1,2]:= 0;M[1,3]:= 0;M[1,4]:=1;
end;

procedure TDraw3D.AxisSymY(var M:TMatrix);
begin
  M[1,1]:=-1;M[1,2]:=0;M[1,3]:= 0;M[1,4]:=0;
  M[2,1]:= 0;M[2,2]:=1;M[2,3]:= 0;M[2,4]:=0;
  M[3,1]:= 0;M[3,2]:=0;M[3,3]:=-1;M[3,4]:=0;
  M[4,1]:= 0;M[1,2]:=0;M[1,3]:= 0;M[1,4]:=1;
end;

procedure TDraw3D.AxisSymZ(var M:TMatrix);
begin
  M[1,1]:=-1;M[1,2]:= 0;M[1,3]:=0;M[1,4]:=0;
  M[2,1]:= 0;M[2,2]:=-1;M[2,3]:=0;M[2,4]:=0;
  M[3,1]:= 0;M[3,2]:= 0;M[3,3]:=1;M[3,4]:=0;
  M[4,1]:= 0;M[1,2]:= 0;M[1,3]:=0;M[1,4]:=1;
end;

procedure TDraw3D.PlaneSymXY(var M:TMatrix);
begin
  M[1,1]:=1;M[1,2]:=0;M[1,3]:= 0;M[1,4]:=0;
  M[2,1]:=0;M[2,2]:=1;M[2,3]:= 0;M[2,4]:=0;
  M[3,1]:=0;M[3,2]:=0;M[3,3]:=-1;M[3,4]:=0;
  M[4,1]:=0;M[1,2]:=0;M[1,3]:= 0;M[1,4]:=1;
end;

procedure TDraw3D.PlaneSymXZ(var M:TMatrix);
begin
  M[1,1]:=1;M[1,2]:= 0;M[1,3]:=0;M[1,4]:=0;
  M[2,1]:=0;M[2,2]:=-1;M[2,3]:=0;M[2,4]:=0;
  M[3,1]:=0;M[3,2]:= 0;M[3,3]:=1;M[3,4]:=0;
  M[4,1]:=0;M[1,2]:= 0;M[1,3]:=0;M[1,4]:=1;
end;

procedure TDraw3D.PlaneSymYZ(var M:TMatrix);
begin
  M[1,1]:=-1;M[1,2]:=0;M[1,3]:=0;M[1,4]:=0;
  M[2,1]:= 0;M[2,2]:=1;M[2,3]:=0;M[2,4]:=0;
  M[3,1]:= 0;M[3,2]:=0;M[3,3]:=1;M[3,4]:=0;
  M[4,1]:= 0;M[1,2]:=0;M[1,3]:=0;M[1,4]:=1;
end;

procedure TDraw3D.OriginSym(var M:TMatrix);
begin
  M[1,1]:=-1;M[1,2]:= 0;M[1,3]:= 0;M[1,4]:=0;
  M[2,1]:= 0;M[2,2]:=-1;M[2,3]:= 0;M[2,4]:=0;
  M[3,1]:= 0;M[3,2]:= 0;M[3,3]:=-1;M[3,4]:=0;
  M[4,1]:= 0;M[1,2]:= 0;M[1,3]:= 0;M[1,4]:=1;
end;

procedure TDraw3D.GenRot(A,B:T3DPoint;Alfa:Double;var M:TMatrix);
var Mat1,Mat2,Mat       :TMatrix;
    V                   :TVector;
    i                   :Integer;
    Beta,Gama           :Real;
begin
    V[1]:=-A[1];V[2]:=-A[2];V[3]:=-A[3];
    Transl(V,Mat1);
    Beta:=-Arctan((B[2]-A[2])/(B[1]-A[1]));
    RotZ(Beta,Mat2);
    Multip(Mat1,Mat2,Mat);
    Mat1:=Mat;
    Gama:=Arctan(sqrt(sqr(B[1]-A[1])+sqr(B[2]-A[2]))/(B[3]-A[3]));
    RotY(Gama,Mat2);
    Multip(Mat1,Mat2,Mat);
    Mat1:=Mat;
    RotZ(Alfa,Mat2);
    Multip(Mat1,Mat2,Mat);
    Mat1:=Mat;
    Gama:=-Gama;
    RotY(Gama,Mat2);
    Multip(Mat1,Mat2,Mat);
    Mat1:=Mat;
    Beta:=-Beta;
    RotZ(Beta,Mat2);
    Multip(Mat1,Mat2,Mat);
    Mat1:=Mat;
    V[1]:=-V[1];V[2]:=-V[2];V[3]:=-V[3];
    Transl(V,Mat2);
    Multip(Mat1,Mat2,Mat);
    M:=Mat;
end;

Function Prvek(Zapis:String;i:Integer; var Hodnota:Double; var Znak:Char;var PocetZnaku:Integer):Boolean;
//Funkce pro ten zpisu po prvcch.
//vrac true, pokud se jet etl znak
//vrac znak "$" pokud se etlo slo

var j,Code    :Integer;
begin
    if i>Length(zapis)
          then Prvek:=False
          else begin
                 Prvek:=True;
                 if Zapis[i] in digits
                 then begin
                           Znak:='$'; {na vstupu je slo}
                           j:=i;
                           Repeat
                             inc(i);
                             if Zapis[i]='E' then inc(i);
                             if ((Zapis[i]='+') or (Zapis[i]='-')) and (Zapis[pred(i)]='E')
                                             then inc(i);
                           Until not(Zapis[i] in Digits);
                           PocetZnaku:=i-j;
                           Val(Copy(Zapis,j,i-j),Hodnota,Code);
                           if Code>0 then begin Znak:='#';exit;end;
                        end
                    else Znak:=Zapis[i]; {Precte jiny znak}
               end;
end;

//Pomocn funkce pro vyslen etezce

Function FunctionReplacement(Vstup:String):String;
// funkce zamuje zpisy matematickch funkc za jednopismenove priznaky
var i,j,k,Left,Right :Integer;
    Vystup,Vypis     :String;
begin
  for i:=1 to length(Vstup) do Vstup[i]:=UpCase(Vstup[i]);

   i:=0;
  Repeat
   inc(i);
   if Vstup[i]=',' then Vstup[i]:='.';
   if not (Vstup[i] in Letters+BinOper+Digits+['(',')']) then
     begin
       FunctionReplacement:='Chybny vyraz';exit;
     end;
  until i=length(Vstup);

  Left:=0;Right:=0;

  i:=0;
  Repeat
     inc(i);
     Case Vstup[i] of
    '.'                 : if (not (Vstup[pred(i)] in Digits)) or
                             (not (Vstup[succ(i)] in Digits)) then
                           begin
                              FunctionReplacement:='Chybny vyraz';exit;
                           end;
    '('                 :inc(Left);
    ')'                 :inc(Right);
    '^','*','/','+','-' :if (Vstup[succ(i)] in binoper+[')']) then
                         begin
                           FunctionReplacement:='Chybny vyraz';exit;
                         end;
    'A','E','F','S','N','R','T','L','C',
    'K','X','P','Y','Z':begin
                          Vypis:=Vstup[i];
                          j:=succ(i);
                          While Vstup[j] in Letters do
                            begin Vypis:=Vypis+Vstup[j];inc(j);end;
                          if j=succ(i)
                            then begin
                                  if (not (Vstup[i] in Variables))
                                   then
                                    begin
                                      FunctionReplacement:='Chybny vyraz';exit;
                                    end;
                                 end
                            else begin
                             k:=0;
                             Repeat
                               inc(k);
                             until (Vypis=Functions[k]) or (k=pred(NoOfFunct));
                             if Vypis<>'PI' then
                             if (Vypis<>Functions[k]) or (Vstup[j]<>'(')
                             then begin
                                    FunctionReplacement:='Chybny vyraz';exit;
                                  end
                          end;
                          i:=pred(j);
                         end;
      '0'..'9'         :begin
                        end;
      else begin
            FunctionReplacement:='Chybny vyraz';exit;
           end;
    end;
 until i=Length(Vstup);
 if Left<>Right then
      begin
       FunctionReplacement:='Chybny vyraz';exit;
     end;

  Vystup:='';
  i:=1;
  Repeat
       Case Vstup[i] of
        {funkce}
       '-':  begin
             if (i=1) or (Vstup[i-1]='(')
                                      then Vystup:=Vystup + #199  {unrn minus}
                                      else Vystup:=Vystup + Vstup[i];
              inc(i);
             end;
       '+':  begin
             if (i=1) or (Vstup[i-1]='(')
                                      then Vystup:=Vystup          {unrn plus}
                                      else Vystup:=Vystup + Vstup[i];
              inc(i);
             end;
       'S': Case Vstup[i+2] of
               'N':begin Vystup:=Vystup + #200;                          {sinus}
                         i:=i+3;
                   end;
               'R':begin
                   if Vstup[i+3]='T'
                   then begin Vystup:=Vystup + #207;i:=i+4;end {druha odmocnina}
                   else begin Vystup:=Vystup + #206;i:=i+3;end   {druha mocnina}
                   end;
               else begin Vystup:=Vystup+'S';inc(i);end;            {promnn S}
               end;
       'C':if Vstup[succ(i)]='O'
               then begin Vystup:=Vystup + #201;i:=i+3;end;            {cosinus}
       'A':Case Vstup[i+3] of
                'T':begin Vystup:=Vystup + #204;i:=i+6;end;             {arctan}
                'S':begin Vystup:=Vystup + #209;i:=i+6;end;             {arcsin}
                'C':begin Vystup:=Vystup + #210;i:=i+6;end              {arccos}
                 else begin Vystup:=Vystup + #208;i:=i+3;end;              {abs}
               end;
       'E':begin Vystup:=Vystup + #202;i:=i+3;end;                         {exp}
       'L':begin Vystup:=Vystup + #203;i:=i+2;end;                          {ln}
       'T':if Vstup[succ(i)]='A'
             then begin Vystup:=Vystup + #205;i:=i+3;end                   {tan}
             else begin Vystup:=Vystup+'T';inc(i);end;              {promnn T}
       'F':begin Vystup:=Vystup + #211;i:=i+2;end;              {fa - faktoril}
       'P':begin Vystup:=Vystup+'P';i:=i+2;end;                  {Konstanta pi}
       {Promnn}
       'X':begin Vystup:=Vystup+'X';inc(i);end;                     {promnn X}
       'Y':begin Vystup:=Vystup+'Y';inc(i);end;                     {promnn Y}
       'Z':begin Vystup:=Vystup+'Z';inc(i);end;                     {promnn Z}
       'R':begin Vystup:=Vystup+'R';inc(i);end;                     {promnn R}
       'N':begin Vystup:=Vystup+'N';inc(i);end;                     {promnn N}
       'K':begin Vystup:=Vystup+'K';inc(i);end;                     {promnn K}
        else begin Vystup:=Vystup + Vstup[i];inc(i);end
       end;                                                               {Case}
  Until i>Length(Vstup);
  FunctionReplacement:=Vystup;
end;

Function TDraw3D.PreCalc(Vstup:String;var Errorreport:String):TExpression;
Const Max = 255;                                     {maximalni pocet operatoru}
      NotInit = 0;                                    {neinicializovana hodnota}
var   Zasobnik  : TExpression;                               {Pracovni zasobnik}
      W         : 0..Max;                                     {Vrchol zasobniku}
      H         : Double;                                     {hodnoty operand}
      Z         : Char;                                         {znak na vstupu}
      Pokracovat: Boolean;
      i,j,Pocet : Integer;
      VypisCisla:String;
      Vystup    :TExpression;
      Term      :TTerm;
begin
  VypisCisla:=Vstup;
  Vstup:=FunctionReplacement(Vstup);
  if (vstup='Chybny vyraz') or
     (vstup='Chybne cislo') then
     begin
       ErrorReport:=Vstup;
       exit;
     end;
  For j:=1 to max do
      With Vystup[j] do begin Value:=NotInit;Character:='%';end;
  w:=0;i:=1;j:=0;
  Pokracovat:=Prvek(Vstup,i,H,Z,Pocet);
  While Pokracovat do
  begin
   Case Z of
   '$':With Vystup[j] do
       begin                                                {na vstupu je cislo}
             Value:=H;Character:='$';
             inc(j);
             i:=i+Pocet;
       end;
   '#':begin ErrorReport:='Chybne cislo';                         {chybne cislo}
             exit;
       end;
   'P':With Vystup[j] do
       begin                                                   {na vstupu je pi}
             Value:=Pi;Character:='$';inc(j);inc(i);
       end;
   'K','N','R','S','T','X','Y','Z':
       With Vystup[j] do
       begin                                             {na vstupu je promenna}
             Value:=NotInit;Character:=Z;inc(j);inc(i);
       end;
   '(':begin                                              {zavorku do zasobniku}
             inc(W);
             Term.Character:='(';Term.Value:=NotInit;
             Zasobnik[w]:=Term;inc(i);
       end;
   ')':begin                                  {z vrcholu zasobniku az k zavorce}
         while (w>0) and (Zasobnik[w].Character<>'(') do
           begin  Vystup[j]:=Zasobnik[w];inc(j);dec(w) end;
           if w=0 then begin Errorreport:='Chybny vyraz';exit;end
                  else dec(w);                                 {zruseni zavorky}
         inc(i);
       end;
 ' ' :inc(i);
 #200..#211:                    {operatory nejvyssi priority - ven ze zasobniku}
       begin
         while (w>0) and (Zasobnik[w].Character in [#200..#211]) do
             begin
                   Vystup[j]:=Zasobnik[w];
                   inc(j);dec(w) end;
             inc(w);Term.Character:=Z;Term.Value:=NotInit;
             Zasobnik[w]:=Term;                          {znak do zasobniku}
           inc(i);
       end;

 '^' :  begin                                       {mocnina - ven ze zasobniku}
           while
           (w>0) and ((Zasobnik[w].Character='^') or (Zasobnik[w].Character in [#200..#211])) do
             begin Vystup[j]:=Zasobnik[w];inc(j);dec(w) end;
             inc(w);
             Term.Character:=Z;Term.Value:=NotInit;
             Zasobnik[w]:=Term;                              {znak do zasobniku}
           inc(i);
         end;

 '/','*':begin                     {operatory nizsi priority - ven ze zasobniku}
           while (w>0) and (Zasobnik[w].Character in [#200..#211,'*','/','^']) do
             begin Vystup[j]:=Zasobnik[w];inc(j);dec(w) end;
             inc(w);Term.Character:=Z;Term.Value:=NotInit;
             Zasobnik[w]:=Term;                              {znak do zasobniku}
           inc(i);
         end;
  #199:  begin                                                   {unarni minus}
           while (w>0) and (Zasobnik[w].Character in [#200..#211,'+','-','*','/','^']) do
           begin
             Vystup[j]:=Zasobnik[w];
             inc(j);dec(w)
           end;
             inc(w);Term.Character:=Z;Term.Value:=NotInit;
             Zasobnik[w]:=Term;                             {znak do zasobniku}
             inc(i);
           end;
 '+','-':begin                  {operatory nejnizsi priority - ven ze zasobniku}
           while (w>0) and (Zasobnik[w].Character in [#199..#211,'+','-','*','/','^']) do
             begin Vystup[j]:=Zasobnik[w];inc(j);dec(w) end;
             inc(w);Term.Character:=Z;Term.Value:=NotInit;
             Zasobnik[w]:=Term;                              {znak do zasobniku}
           inc(i);
         end;
 
     end; {Case}
    Pokracovat:=Prvek(Vstup,i,H,Z,Pocet);
  end;
  While (w>0) and (Zasobnik[w].Character<>'(') do
       begin                              {vyprazdnit pripadny zbytek zasobniku}
        Vystup[j]:=Zasobnik[w];inc(j);dec(w);
       end;
  PreCalc:=Vystup;
  if w>0 then Errorreport:='Chybny vyraz'
         else Errorreport:='';
end;                                                       {Procedure PreExpres}

Function TDraw3D.PostCalc(Vstup:TExpression;var ErrorReport:String):Double;
var   Zasobnik,ZZZ  : TExpression;                 {zsobnk na hodnoty}
      i,w       : Integer;
      H,H1,H2   :Double;

Function Faktorial(N:LongInt):LongInt;
begin
   if n>1 then Faktorial:=n*Faktorial(pred(n))
          else Faktorial:=1;
end;

Function Mocnina(Zaklad:Double;Exponent:LongInt):Double;
var i:LongInt;
    Moc:Double;
begin
  Moc:=1;
  for i:=1 to abs(Exponent) do Moc:=Moc*Zaklad;
  if Exponent<0 then Moc:=1/Moc;
  Mocnina:=Moc;
end;

begin
  if (ErrorReport='Chybny vyraz')
      or (ErrorReport='Chybne cislo') then Exit;
  ErrorReport:='';
  w:=-1;i:=0;
  While (Vstup[i].Character<>'%') do
  begin
    Case Vstup[i].Character of
   '$':begin inc(w);                                        {na vstupu je cislo}
             Zasobnik[w]:=Vstup[i];
             inc(i);
       end;
   'K','N','R','S','T','X','Y','Z':                      {na vstupu je promenna}
       begin
            inc(w);
            Case Vstup[i].Character of
            'K': Vstup[i].Value:=K;
            'N': Vstup[i].Value:=N;
            'R': Vstup[i].Value:=R;
            'S': Vstup[i].Value:=S;
            'T': Vstup[i].Value:=T;
            'X': Vstup[i].Value:=X;
            'Y': Vstup[i].Value:=Y;
            'Z': Vstup[i].Value:=Z;
            end;
           Zasobnik[w]:=Vstup[i];
           inc(i);
       end;
     '^','/','*', '+','-':             {binrn opertory}
        begin
          H1:=Zasobnik[w].Value;
          dec(w);
          H2:=Zasobnik[w].Value;
           Case Vstup[i].Character of
            '^':begin
                 {H2^H1}
                 if (Abs(H2)<PractZero) and (H1<0) then
                    begin
                      ErrorReport:='Neppustn zklad reln mocniny';
                      exit;
                    end
                 else   
                 if (abs(H1)<=0) and (H1=Round(h1))
                          then Zasobnik[w].Value:=Mocnina(H2,Round(H1))
                     else if H1=Round(h1)
                        then Zasobnik[w].Value:=Mocnina(H2,Round(H1))
                          else if h2>PractZero
                              then Zasobnik[w].Value:=exp(H1*ln(H2))
                              else begin
                                    ErrorReport:='Neppustn zklad reln mocniny';
                                    exit;
                                   end;
                end;
            '/':begin
                  if Abs(H1)<PractZero
                   then begin
                          ErrorReport:='Dlen nulou';
                          exit;
                        end
                        else Zasobnik[w].Value:=H2 /   H1;
                end;
            '*':Zasobnik[w].Value:=H2 *   H1;
            '+':Zasobnik[w].Value:=H2 +   H1;
            '-':Zasobnik[w].Value:=H2 -   H1;
            end;
            inc(i);
          end;
                                           {funkce}
        #199:begin
              inc(i);Zasobnik[w].Value:=-Zasobnik[w].Value;      {unrn minus}
             end;
        #200:begin
              inc(i);Zasobnik[w].Value:=sin(Zasobnik[w].Value);          {sinus}
             end;
        #201:begin
              inc(i);Zasobnik[w].Value:=cos(Zasobnik[w].Value);        {cosinus}
             end;
        #202:begin
              inc(i);Zasobnik[w].Value:=exp(Zasobnik[w].Value);            {exp}
             end;
        #203:begin
              inc(i);
              if Zasobnik[w].Value>PractZero
                          then Zasobnik[w].Value:=ln(Zasobnik[w].Value)     {ln}
                          else begin
                                ErrorReport:='Neppustn argument logaritmu';
                                exit;
                               end;
             end;
        #204:begin
              inc(i);Zasobnik[w].Value:=arctan(Zasobnik[w].Value);      {arctan}
             end;
        #205:begin
              inc(i);                                                  {tangens}
              if abs(cos(Zasobnik[w].Value))>PractZero
                            then Zasobnik[w].Value:=sin(Zasobnik[w].Value)/cos(Zasobnik[w].Value)
                            else begin
                                  ErrorReport:='Nedefinovan tangens';
                                  exit;
                                 end;
             end;
        #206:begin
              inc(i);Zasobnik[w].Value:=sqr(Zasobnik[w].Value);{druh mocnina}
             end;
        #207:begin
              inc(i);                                          {druh odmocnina}
              if Zasobnik[w].Value>=0
                      then Zasobnik[w].Value:=sqrt(Zasobnik[w].Value)
                      else begin
                           ErrorReport:='Zporn argument odmocniny';
                           exit;
                           end;
             end;
        #208:begin
              inc(i);Zasobnik[w].Value:=Abs(Zasobnik[w].Value);{absolutn hodnota}
             end;
        #209:begin                                                 {arcus sinus}
              inc(i);
              if abs(Zasobnik[w].Value)<1
                then Zasobnik[w].Value:=
                 arctan(Zasobnik[w].Value/sqrt(1-Zasobnik[w].Value*Zasobnik[w].Value))
                else if Zasobnik[w].Value=-1
                    then Zasobnik[w].Value:=-pi/2
                    else if Zasobnik[w].Value=1
                        then Zasobnik[w].Value:=pi/2
                        else begin
                              ErrorReport:='Neppustn argument arcsin';
                              exit;
                             end;
             end;
        #210:begin                                               {arcus cosinus}
              inc(i);
              if abs(Zasobnik[w].Value)<1
                then Zasobnik[w].Value:=pi/2-arctan(Zasobnik[w].Value/sqrt(1-Zasobnik[w].Value*Zasobnik[w].Value))
                else if Zasobnik[w].Value=-1
                    then Zasobnik[w].Value:=pi
                    else if Zasobnik[w].Value=1
                        then Zasobnik[w].Value:=0
                        else begin
                              ErrorReport:='Neppustn argument arccos';
                              exit;
                             end;
             end;
        #211:begin                                                   {faktoril}
               inc(i);
               if (Zasobnik[w].Value>=0) and (Round(Zasobnik[w].Value)=Zasobnik[w].Value)
                  then Zasobnik[w].Value:=Faktorial(Round(Zasobnik[w].Value))
                  else begin
                        ErrorReport:='Neppustn argument faktorilu';
                        exit;
                       end;
             end;
         end;                                                           {Case Z}
        end;
  if w>0 then ErrorReport:='Chybn vraz';
  PostCalc:=Zasobnik[w].Value;
end;                                                                {PostExpres}

Function TDraw3D.Calc(Vstup:String;var ErrorReport:String):Double;
var Expression:TExpression;
begin
  Expression:=PreCalc(Vstup,ErrorReport);
  Calc:=PostCalc(Expression,ErrorReport);
end;


end.
